#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
CloudFiler.py

Main GUI script.  For a description of what CloudFiler does see:

http://SF/cloud-filer-backup-tool.html

The icons used in Cloud Filer are open source and from the following sources:

Page-world, Film, Lock, Package, Image, Word, Excel & Powerpoint from
    http://findicons.com/pack/1588/farm_fresh_web by FatCow Web Hosting
    (http://creativecommons.org/licenses/by/3.0/)

Others from http://openiconlibrary.sourceforge.net/
    (Either public domain or http://creativecommons.org/licenses/by-sa/1.0/)
'''

import wx
import sys
import os
import time
import re
import fnmatch
import datetime
import encodings.hex_codec
import encodings.ascii
import encodings.utf_8
import encodings.idna
import threading
import imp
import passlib.handlers.pbkdf2
import httplib
import Utils

from time import sleep
from itertools import chain
from FileListCtrl import FileListCtrl

import NullLicensing as Licensing
#import Licensing
import AppConfig

ID_LOCAL_DIR_COMBO = 100
ID_REMOTE_DIR_COMBO = 101
ID_REMOTE_CONNECT = 102
ID_CLOUD_COMBO = 103
ID_NEW_CLOUD = 106
ID_DELETE_CLOUD = 107
ID_OPEN_HELP = 108
ID_OPEN_HELP_ABOUT = 109
ID_LOCAL_UPLOAD = 110
ID_REMOTE_DOWNLOAD = 111
ID_CLOUD_ENCRYPT = 112
ID_OPEN_HELP_LICENSING = 113
ID_FILE_SETTINGS = 114

def DEBUG(msg):
# Uncomment the following line for debug output
    print msg.encode('utf-8', errors='replace')
#    pass


ID_OS_LINUX = 1
ID_OS_MACOS = 2
ID_OS_WINDOWS = 3

def GetOS():
    if os.name == 'posix':
        # Either Linux or MacOS
        DEBUG('sys.platform=' + sys.platform)
        if sys.platform.lower().startswith('linux'):
            return ID_OS_LINUX
        else:
            return ID_OS_MACOS
    else:
        return ID_OS_WINDOWS


CHUNK_SIZE = 8 * 1024 * 1024

CF_VERSION_STRING = '1.3'



class NewBucketOrFolderDlg(wx.Dialog):

    def __init__(self, isBucket, *args, **kwargs):
        super(NewBucketOrFolderDlg, self).__init__(*args, **kwargs) 
        
        self.InitUI(isBucket)

        
    def InitUI(self, isBucket):
        self.SetSize((400, 130))

        if isBucket:
            labelText = 'Bucket name:'
        else:
            labelText = 'Folder name:'
        label1 = wx.StaticText(self, label=labelText, pos=(10, 15))
        self.newName = wx.TextCtrl(self, pos=(106, 11))
        self.newName.SetFocus()
        cw = self.newName.GetCharWidth()
        w,h = self.newName.GetSizeTuple()
        self.newName.SetSize((cw * 32, h))

        okButton = wx.Button(self, id=wx.ID_OK, label='Ok', pos=(200, 65))
        okButton.Bind(wx.EVT_BUTTON, self.OnClose)
        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(300, 65))

    def OnClose(self, e):        
        # Validate fields
        newName = self.newName.GetValue()
        if newName.startswith('_CF'):
            dlg = wx.MessageDialog(None, 'Invalid name (cannot start with _CF)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif len(newName) < 1:
            dlg = wx.MessageDialog(None, 'Empty name', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        else:
            self.EndModal(wx.ID_OK)

    def GetName(self):
        return self.newName.GetValue()


class GetNewLicenseDlg(wx.Dialog):

    def __init__(self, licMsg, *args, **kwargs):
        super(GetNewLicenseDlg, self).__init__(*args, **kwargs) 
        
        self.InitUI(licMsg)

        
    def InitUI(self, licMsg):
        self.SetSize((400, 130))

        label0 = wx.StaticText(self, label=licMsg, pos=(10, 15))

        label1 = wx.StaticText(self, label='License Key:', pos=(10, 45))
        self.licenseKey = wx.TextCtrl(self, pos=(106, 41))
        self.licenseKey.SetFocus()
        cw = self.licenseKey.GetCharWidth()
        w,h = self.licenseKey.GetSizeTuple()
        self.licenseKey.SetSize((cw * 32, h))

        okButton = wx.Button(self, id=wx.ID_OK, label='Ok', pos=(200, 75))
        okButton.Bind(wx.EVT_BUTTON, self.OnClose)
        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(300, 75))

    def OnClose(self, e):        
        # Validate fields
        licenseKey = self.licenseKey.GetValue()
        if len(licenseKey) < 1:
            dlg = wx.MessageDialog(None, 'Empty key', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        else:
            self.EndModal(wx.ID_OK)

    def GetKey(self):
        return Utils.ensure_ascii(self.licenseKey.GetValue().strip())



class RenameDlg(wx.Dialog):

    def __init__(self, oldName, *args, **kwargs):
        super(RenameDlg, self).__init__(*args, **kwargs) 
        
        self.InitUI(oldName)

        
    def InitUI(self, oldName):
        self.SetSize((400, 180))

        label1 = wx.StaticText(self, label='Old name:', pos=(10, 15))
        oldNameText = wx.StaticText(self, label=oldName, pos=(110, 15))

        label1 = wx.StaticText(self, label='New name:', pos=(10, 45))
        self.newName = wx.TextCtrl(self, pos=(106, 41))
        self.newName.SetValue(oldName)
        self.newName.SelectAll()
        self.newName.SetFocus()
        cw = self.newName.GetCharWidth()
        w,h = self.newName.GetSizeTuple()
        self.newName.SetSize((cw * 32, h))

        okButton = wx.Button(self, id=wx.ID_OK, label='Ok', pos=(200, 85))
        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(300, 85))

    def GetNewName(self):
        return self.newName.GetValue()


class NewCloudDlg(wx.Dialog):

    def __init__(self, *args, **kwargs):
        super(NewCloudDlg, self).__init__(*args, **kwargs) 
        self.InitUI()
        
    def InitUI(self):
        self.SetSize((400, 180))

        label1 = wx.StaticText(self, label='Cloud Service:', pos=(10, 15))
        self.cloudService = wx.TextCtrl(self, pos=(130, 10))
        cw = self.cloudService.GetCharWidth()
        w,h = self.cloudService.GetSizeTuple()
        self.cloudService.SetSize((cw * 22, h))

        label2 = wx.StaticText(self, label='Cloud Host:', pos=(10, 45))
        self.cloudHost = wx.TextCtrl(self, pos=(130, 40))
        cw = self.cloudHost.GetCharWidth()
        w,h = self.cloudHost.GetSizeTuple()
        self.cloudHost.SetSize((cw * 32, h))

        label3 = wx.StaticText(self, label='Short Name:', pos=(10, 75))
        self.shortName = wx.TextCtrl(self, pos=(130, 70))
        cw = self.shortName.GetCharWidth()
        w,h = self.shortName.GetSizeTuple()
        self.shortName.SetSize((cw * 4, h))

        okButton = wx.Button(self, id=wx.ID_OK, label='Ok', pos=(200, 105))
        okButton.Bind(wx.EVT_BUTTON, self.OnClose)
        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(300, 105))


    def OnClose(self, e):        
        # Validate fields
        cloudName = self.cloudService.GetValue()
        cloudHost = self.cloudHost.GetValue()
        shortName = self.shortName.GetValue().upper()
        if '/' in cloudName or '\\' in cloudName or ':' in cloudName:
            dlg = wx.MessageDialog(None, 'Invalid cloud name (no slashes or colons)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif len(cloudName) < 1:
            dlg = wx.MessageDialog(None, 'Empty cloud name', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif '/' in cloudHost or '\\' in cloudHost or ':' in cloudHost or ' ' in cloudHost:
            dlg = wx.MessageDialog(None, 'Invalid cloud host (no slashes, spaces or colons)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif len(cloudHost) < 1:
            dlg = wx.MessageDialog(None, 'Empty cloud host', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif len(shortName) != 2 or not shortName.isalpha():
            dlg = wx.MessageDialog(None, 'Invalid short name (expect two alphabetic characters)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif AppConfig.HasSection(cloudName):
            dlg = wx.MessageDialog(None, 'Already have a cloud called ' + cloudName, 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif self.alreadyGotShortName(shortName):
            dlg = wx.MessageDialog(None, 'Already have a short name = ' + shortName, 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        else:
            self.EndModal(wx.ID_OK)
            

    def alreadyGotShortName(self, shortName):
        for section in AppConfig.GetSections():
            if section != 'History':
                if AppConfig.Get(section, 'ShortName') == shortName:
                    return True
        
        return False


    def GetCloudService(self):
        return self.cloudService.GetValue()


    def GetCloudHost(self):
        return self.cloudHost.GetValue()


    def GetShortName(self):
        return self.shortName.GetValue().upper()


class CloudConnectDlg(wx.Dialog):

    def __init__(self, *args, **kwargs):
        super(CloudConnectDlg, self).__init__(*args, **kwargs) 
            
        self.InitUI()
        

    def InitUI(self):
        self.SetSize((430, 415)) # Y size = OkButton YPos + 70

        # Cloud connection fields
        label1 = wx.StaticText(self, label='Cloud Service:', pos=(10, 15))
        self.cloudCombo = wx.ComboBox(self, ID_CLOUD_COMBO, style=wx.TE_PROCESS_ENTER, pos=(130, 10)) 
        delButton = wx.Button(self, id=ID_DELETE_CLOUD, label='Delete', pos=(330, 10))
        for section in AppConfig.GetSections():
            if section != 'History':
                self.cloudCombo.Append(section)
        self.cloudCombo.Append('New...')

        label2 = wx.StaticText(self, label='Cloud Host:', pos=(10, 45))
        label3 = wx.StaticText(self, label='Short Name:', pos=(10, 75))

        self.hostText = wx.StaticText(self, pos=(130, 45))
        self.rootText = wx.StaticText(self, pos=(130, 75))

        # Cloud authentication fields
        box1 = wx.StaticBox(self, label='User', pos=(5, 100), size=(410, 115))

        label4 = wx.StaticText(self, label='Access Key:', pos=(10, 125))
        label5 = wx.StaticText(self, label='Secret Key:', pos=(10, 155))

        self.accessKey = wx.TextCtrl(self, style=wx.TE_PASSWORD, pos=(130, 121))
        cw = self.accessKey.GetCharWidth()
        w,h = self.accessKey.GetSizeTuple()
        self.accessKey.SetSize((cw * 35, h))

        self.secretKey = wx.TextCtrl(self, style=wx.TE_PASSWORD, pos=(130, 151))
        cw = self.secretKey.GetCharWidth()
        w,h = self.secretKey.GetSizeTuple()
        self.secretKey.SetSize((cw * 35, h))

        self.storeKeys = wx.CheckBox(self, label='Save keys', pos=(130, 181))

        # Encryption fields
        box2 = wx.StaticBox(self, label='Security', pos=(5, 220), size=(410, 105))
        self.encrypt = wx.CheckBox(self, id=ID_CLOUD_ENCRYPT, label='Encrypt', pos=(130, 235))

        self.passwordLabel = wx.StaticText(self, label='Password:', pos=(10, 265))
        self.passwordLabel.Enable(False)

        self.password = wx.TextCtrl(self, style=wx.TE_PASSWORD, pos=(130, 261))
        self.password.Enable(False)
        cw = self.password.GetCharWidth()
        w,h = self.password.GetSizeTuple()
        self.password.SetSize((cw * 35, h))

        self.storePassword = wx.CheckBox(self, label='Save password', pos=(130, 291))
        self.storePassword.Enable(False)

        if not Licensing.EncryptionModuleLicensed():
            self.encrypt.Enable(False)

        # Bottom buttons
        okButton = wx.Button(self, id=wx.ID_OK, label='Ok', pos=(230, 345))
        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(330, 345))

        # Wire up events
        self.Bind(wx.EVT_COMBOBOX, self.OnChangeCloud, id=ID_CLOUD_COMBO)
        self.Bind(wx.EVT_BUTTON, self.OnNewCloud, id=ID_NEW_CLOUD)
        self.Bind(wx.EVT_BUTTON, self.OnDeleteCloud, id=ID_DELETE_CLOUD)
        self.Bind(wx.EVT_CHECKBOX, self.OnChangeEncrypt, id=ID_CLOUD_ENCRYPT)
        okButton.Bind(wx.EVT_BUTTON, self.OnClose)

        # Set the combobox selection
        selectedCloud = AppConfig.Get('History', 'LastCloud')
        try:
            sn = AppConfig.Get(selectedCloud, 'ShortName')
            if len(sn) == 0:
                raise Exception('That doesnt look any good to me')
        except:
            selectedCloud = AppConfig.GetSections()[0] # Didn't find that cloud so use the first
        self.SetCloudSelection(selectedCloud)

        okButton.SetFocus()


    def OnClose(self, e):        
        # Validate fields
        if self.encrypt.GetValue() and len(self.password.GetValue()) < 4:
            dlg = wx.MessageDialog(None, 'Invalid password (must be at least 4 characters long)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif len(self.accessKey.GetValue()) < 1:
            dlg = wx.MessageDialog(None, 'Invalid access key (must be at least 1 character long)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        elif len(self.secretKey.GetValue()) < 1:
            dlg = wx.MessageDialog(None, 'Invalid secret key (must be at least 1 character long)', 'Error', wx.OK | wx.ICON_ERROR)
            dlg.ShowModal()
        else:
            self.EndModal(wx.ID_OK)


    def OnChangeEncrypt(self, e):
        if self.encrypt.GetValue() and Licensing.EncryptionModuleLicensed():
            self.password.Enable(True)
            self.passwordLabel.Enable(True)
            self.storePassword.Enable(True)
            try:
                suppressWarn = AppConfig.Get('History', 'UserSupressedPasswordWarning')
            except:
                suppressWarn = 'N'
            if suppressWarn != 'Y':
                dlg = Utils.InformationDlg(Utils.ACTION_SWITCH_ENCRYPTED, True, self, wx.ID_ANY, 'Cloud Filer - Warning')
                dlg.ShowModal()
                if dlg.GetEnoughAlready():
                    AppConfig.Set('History', 'UserSupressedPasswordWarning', 'Y')
                    AppConfig.WriteSettings()
                dlg.Destroy()

        else:
            self.password.Enable(False)
            self.passwordLabel.Enable(False)
            self.storePassword.Enable(False)


    def SetCloudSelection(self, selection):
        self.cloudCombo.SetStringSelection(selection)

        try:
            cloudHost = AppConfig.Get(selection, 'Host')
        except:
            cloudHost = ''

        try:
            shortName = AppConfig.Get(selection, 'ShortName')
        except:
            shortName = ''

        try:
            saveKeys = AppConfig.Get(selection, 'SaveKeys')
            if saveKeys == 'Y':
                self.storeKeys.SetValue(True)
            else:
                self.storeKeys.SetValue(False)
        except:
            self.storeKeys.SetValue(False)

        try:
            encrypt = AppConfig.Get(selection, 'Encrypt')
            if encrypt == 'Y' and Licensing.EncryptionModuleLicensed():
                self.encrypt.SetValue(True)
            else:
                self.encrypt.SetValue(False)
        except:
            self.encrypt.SetValue(False)
        if self.encrypt.GetValue():
            DEBUG('encrypt=Y')
            self.password.Enable(True)
            self.passwordLabel.Enable(True)
            self.storePassword.Enable(True)
        else:
            DEBUG('encrypt=N')
            self.password.Enable(False)
            self.passwordLabel.Enable(False)
            self.storePassword.Enable(False)

        try:
            storePassword = AppConfig.Get(selection, 'StorePassword')
            if storePassword == 'Y' and Licensing.EncryptionModuleLicensed():
                self.storePassword.SetValue(True)
            else:
                self.storePassword.SetValue(False)
        except:
            self.storePassword.SetValue(False)
        if self.storePassword.GetValue():
            DEBUG('storePassword=Y')
        else:
            DEBUG('storePassword=N')

        self.hostText.ClearBackground()
        self.hostText.SetLabel(cloudHost)
        self.rootText.ClearBackground()
        self.rootText.SetLabel(shortName + ':')

        keyKey = 'nuvovis.com CloudFiler ' + AppConfig.Get(selection, 'ShortName')
        DEBUG('keyKey=' + keyKey)
        pw = AppConfig.SecureGet(keyKey, 'AK')
        DEBUG('AK=' + pw)
        self.accessKey.SetValue(pw)

        pw = AppConfig.SecureGet(keyKey, 'SK')
        DEBUG('SK=' + pw)
        self.secretKey.SetValue(pw)

        if Licensing.EncryptionModuleLicensed():
            pw = AppConfig.SecureGet(keyKey, 'PWD')
        else:
            pw = ''
        DEBUG('PWD=' + pw)
        self.password.SetValue(pw)


    def SaveLastConnection(self):
        newSection = 'History'
        if not AppConfig.HasSection(newSection):
            AppConfig.AddSection(newSection)
        cloudService = self.cloudCombo.GetStringSelection()
        DEBUG('cloudService=' + cloudService)
        if self.storeKeys.IsChecked():
            saveKeys = 'Y'
        else:
            saveKeys = 'N'
        DEBUG('saveKeys=' + saveKeys)
        if self.encrypt.IsChecked():
            encrypt = 'Y'
        else:
            encrypt = 'N'
        DEBUG('encrypt=' + encrypt)
        if self.storePassword.IsChecked():
            storePassword = 'Y'
        else:
            storePassword = 'N'
        DEBUG('storePassword=' + storePassword)

        AppConfig.Set(newSection, 'LastCloud', cloudService)
        AppConfig.Set(cloudService, 'SaveKeys', saveKeys)
        AppConfig.Set(cloudService, 'Encrypt', encrypt)
        AppConfig.Set(cloudService, 'StorePassword', storePassword)
        AppConfig.WriteSettings()


    def PopupNewCloudDlg(self):
        dlg = NewCloudDlg(self, wx.ID_ANY, 'Add new cloud')
        if dlg.ShowModal() == wx.ID_OK:
            newSection = dlg.GetCloudService()
            AppConfig.AddSection(newSection)
            AppConfig.Set(newSection, 'Host', dlg.GetCloudHost())
            AppConfig.Set(newSection, 'ShortName', dlg.GetShortName())
            AppConfig.WriteSettings()
            self.cloudCombo.Insert(newSection, self.cloudCombo.GetCount()-1)
            self.SetCloudSelection(newSection)
        else:
            DEBUG('TODO: select the first one')

        dlg.Destroy()


    def OnDeleteCloud(self, e):
        oldCloud = self.cloudCombo.GetValue()
        if oldCloud != 'New...':
            action = Utils.ACTION_DELETE_CLOUD
            dlg = Utils.WarningDlg(action, False, oldCloud, self, wx.ID_ANY, 'Cloud Filer - Warning')
            dlg.ShowModal()
            status = dlg.GetEndId()
            dlg.Destroy()
            if status == wx.ID_OK:
                AppConfig.DelSection(oldCloud)
                AppConfig.WriteSettings()
                self.cloudCombo.Delete(self.cloudCombo.FindString(oldCloud))
                for section in AppConfig.GetSections():
                    if section != 'History':
                        self.SetCloudSelection(section)

    def OnNewCloud(self, e):
        self.PopupNewCloudDlg()

    def OnChangeCloud(self, e):
        newCloud = self.cloudCombo.GetValue()
        if newCloud == 'New...':
            self.PopupNewCloudDlg()
        else:
            self.SetCloudSelection(newCloud)


    def GetAccessKey(self):
        return self.accessKey.GetValue()

    def GetSecretKey(self):
        return self.secretKey.GetValue()

    def GetEncrypt(self):
        return self.encrypt.IsChecked()

    def GetPassword(self):
        return self.password.GetValue()

    def StoreKeys(self):
        return self.storeKeys.GetValue()

    def StorePassword(self):
        return self.storePassword.GetValue()


    def GetCloudService(self):
        return self.cloudCombo.GetValue()


    def GetCloudHost(self):
        return self.hostText.GetLabelText()


    def GetRootText(self):
        return self.rootText.GetLabelText()


class SettingsDlg(wx.Dialog):

    def __init__(self, *args, **kwargs):
        super(SettingsDlg, self).__init__(*args, **kwargs) 
        
        self.InitUI()

        
    def InitUI(self):
        try:
            currentProxyServer = AppConfig.Get('History', 'ProxyServer')
            currentProxyUsername = AppConfig.Get('History', 'ProxyUsername')
            currentProxyPassword = AppConfig.Get('History', 'ProxyPassword')
        except:
            currentProxyServer = ''
            currentProxyUsername = ''
            currentProxyPassword = ''

        self.SetSize((400, 180))

        label1 = wx.StaticText(self, label='Proxy Server:', pos=(10, 15))
        self.proxyServer = wx.TextCtrl(self, pos=(130, 11))
        self.proxyServer.SetFocus()
        self.proxyServer.SetValue(currentProxyServer)
        cw = self.proxyServer.GetCharWidth()
        w,h = self.proxyServer.GetSizeTuple()
        self.proxyServer.SetSize((cw * 32, h))

        label2 = wx.StaticText(self, label='Proxy Username:', pos=(10, 45))
        self.proxyUsername = wx.TextCtrl(self, pos=(130, 41))
        self.proxyUsername.SetValue(currentProxyUsername)
        cw = self.proxyUsername.GetCharWidth()
        w,h = self.proxyUsername.GetSizeTuple()
        self.proxyUsername.SetSize((cw * 32, h))

        label3 = wx.StaticText(self, label='Proxy Password:', pos=(10, 75))
        self.proxyPassword = wx.TextCtrl(self, pos=(130, 71))
        self.proxyPassword.SetValue(currentProxyPassword)
        cw = self.proxyPassword.GetCharWidth()
        w,h = self.proxyPassword.GetSizeTuple()
        self.proxyPassword.SetSize((cw * 32, h))

        okButton = wx.Button(self, id=wx.ID_OK, label='Ok', pos=(200, 115))
        okButton.Bind(wx.EVT_BUTTON, self.OnClose)
        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(300, 115))

    def GetProxyServer(self):
        return self.proxyServer.GetValue()

    def GetProxyUsername(self):
        return self.proxyUsername.GetValue()

    def GetProxyPassword(self):
        return self.proxyPassword.GetValue()

    def OnClose(self, e):        
        # Validate fields
        proxyServer = self.proxyServer.GetValue()
        if len(proxyServer) > 0:
            parts = proxyServer.split(":")
            if len(parts) != 2:
                dlg = wx.MessageDialog(None, 'Invalid proxy setting: format should be server:port', 'Error', wx.OK | wx.ICON_ERROR)
                dlg.ShowModal()
            else:
                self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_OK)

class OutputList(wx.ListBox):
    
    def __init__(self, *args, **kwargs):
        super(OutputList, self).__init__(*args, **kwargs) 
            
        self.InitUI()

        
    def InitUI(self):
        self.SetFont(wx.Font(10, wx.FONTFAMILY_TELETYPE, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL))
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)


    def Append(self, msg):
        rows = msg.split('\n')
        for row in rows:
            index = wx.ListBox.Append(self, row)
        self.SetFirstItem(index)


    def OnRightDown(self, e):
        self.outputListMenu = wx.Menu()
        mi = wx.MenuItem(self.outputListMenu, wx.NewId(), 'Select All')
        self.outputListMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnOutputListSelectAll, mi)
        mi = wx.MenuItem(self.outputListMenu, wx.NewId(), 'Copy')
        self.outputListMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnOutputListCopy, mi)

        self.PopupMenu(self.outputListMenu, e.GetPosition())


    def OnOutputListSelectAll(self, e):
        for i in range(self.GetCount()):
            self.SetSelection(i)


    def OnOutputListCopy(self, e):
        text = ''
        selections = self.GetSelections()
        for i in selections:
            text = text + self.GetString(i) + '\n'

        if wx.TheClipboard.Open():
            wx.TheClipboard.Clear()
            wx.TheClipboard.SetData(wx.TextDataObject(text))
            wx.TheClipboard.Close()



class MainForm(wx.Frame):
    
    def __init__(self, *args, **kwargs):
        super(MainForm, self).__init__(*args, **kwargs) 
            
        self.accessKey = None
        self.secretKey = None
        self.password = None
        
        Utils.set_main_dir() # Find out where we are installed so we can pick up stuff like icon images

        AppConfig.Init('..') # Initial ini file is in parent dir

        # Initialise proxy settings
        try:
            self.proxyServer = AppConfig.Get('History', 'ProxyServer')
            self.proxyUsername = AppConfig.Get('History', 'ProxyUsername')
            self.proxyPassword = AppConfig.Get('History', 'ProxyPassword')
        except:
            self.proxyServer = ''
            self.proxyUsername = ''
            self.proxyPassword = ''

        if self.CheckLicensed(self.proxyServer, self.proxyUsername, self.proxyPassword):
            self.InitUI()

        
    def InitUI(self):

        self.icon = wx.Icon(Utils.GetFilePath(Utils.ICON_APPLICATION), wx.BITMAP_TYPE_ICO)
        self.SetIcon(self.icon)
        
#        icon = wx.IconFromBitmap(wx.Bitmap(GetIconPath(Utils.ICON_REMOTE_FOLDER)))
#        self.SetIcon(icon)#, TRAY_TOOLTIP)

#        image = wx.Image(GetIconPath(Utils.ICON_APPLICATION), wx.BITMAP_TYPE_PNG).ConvertToBitmap()
#        icon = wx.EmptyIcon()
#        icon.CopyFromBitmap(image)
#        self.SetIcon(icon)

#        ib = wx.IconBundle()
#        ib.AddIconFromFile(GetIconPath(Utils.ICON_REMOTE_FOLDER), wx.BITMAP_TYPE_PNG)
#        self.SetIcons(ib)
        
        menubar = wx.MenuBar()

        fileMenu = wx.Menu()
        fileMenu.Append(wx.ID_OPEN, '&Open')
        fileMenu.Append(ID_REMOTE_CONNECT, '&Connect')
        fileMenu.Append(ID_FILE_SETTINGS, '&Settings')
        fileMenu.AppendSeparator()

        qmi = wx.MenuItem(fileMenu, wx.ID_EXIT, '&Quit\tCtrl+W')
        fileMenu.AppendItem(qmi)

        self.Bind(wx.EVT_MENU, self.OnQuit, qmi)

        transferMenu = wx.Menu()
        transferMenu.Append(ID_LOCAL_UPLOAD, 'Upload')
        transferMenu.Append(ID_REMOTE_DOWNLOAD, 'Download')

        helpMenu = wx.Menu()
        helpMenu.Append(ID_OPEN_HELP, 'User Guide')
        helpMenu.Append(ID_OPEN_HELP_LICENSING, 'Licensing')
        helpMenu.Append(ID_OPEN_HELP_ABOUT, 'About')

        menubar.Append(fileMenu, '&File')
        menubar.Append(transferMenu, '&Transfer')
        menubar.Append(helpMenu, '&Help')
        self.SetMenuBar(menubar)

        toolbar = self.CreateToolBar()
        toolbar.SetBackgroundColour('#E0E0E0')
        qtool = toolbar.AddLabelTool(wx.ID_OPEN, 'Open', wx.Bitmap(Utils.GetFilePath(Utils.ICON_FOLDER)), shortHelp='Open')
        qtool = toolbar.AddLabelTool(ID_REMOTE_CONNECT, 'Connect', wx.Bitmap(Utils.GetFilePath(Utils.ICON_REMOTE_FOLDER)), shortHelp='Connect')
        toolbar.AddSeparator()
        qtool = toolbar.AddLabelTool(ID_OPEN_HELP, 'Help', wx.Bitmap(Utils.GetFilePath(Utils.ICON_HELP_CONTENTS)), shortHelp='Help')
        qtool = toolbar.AddLabelTool(ID_OPEN_HELP_ABOUT, 'About', wx.Bitmap(Utils.GetFilePath(Utils.ICON_HELP_ABOUT)), shortHelp='About')
        toolbar.Realize()

        # Create the outer horizontal splitter with a vertical splitter above
        # and the output text box below

        self.hsplitter = wx.SplitterWindow(self, -1)
        self.hsplitter.SetMinimumPaneSize(10)

        # Create the top half of the horizontal splitter
        self.vsplitter = wx.SplitterWindow(self.hsplitter, -1)
        self.vsplitter.SetMinimumPaneSize(10)

        # Create the top left
        topLeftPanel = wx.Panel(self.vsplitter, -1)
        topLeftPanel.SetBackgroundColour('#E0E0E0')
        topRightPanel = wx.Panel(self.vsplitter, -1)
        topRightPanel.SetBackgroundColour('#E0E0E0')

        self.localDirCombo = wx.ComboBox(topLeftPanel, ID_LOCAL_DIR_COMBO, style=wx.TE_PROCESS_ENTER)
#        self.localDirCombo.SetStringSelection(homeDirPath)

        self.localFileList = FileListCtrl(topLeftPanel, -1, style=wx.LC_REPORT)

        topLeftSizer = wx.BoxSizer(wx.VERTICAL)
        topLeftSizer.Add(self.localDirCombo, 0, wx.EXPAND, 1)
        topLeftSizer.Add(self.localFileList, 1, wx.EXPAND | wx.ALL, 1)
        topLeftPanel.SetSizer(topLeftSizer)


        # Create the top right
        self.remoteIcon = wx.StaticBitmap(topRightPanel, bitmap=wx.Bitmap(Utils.GetFilePath(Utils.ICON_RED_SHIELD)), size=(16,16))
        self.remoteDirCombo = wx.ComboBox(topRightPanel, ID_REMOTE_DIR_COMBO, style=wx.TE_PROCESS_ENTER|wx.TE_READONLY)

        topRightIconComboSizer = wx.BoxSizer(wx.HORIZONTAL)
        topRightIconComboSizer.Add(self.remoteIcon, 0, wx.EXPAND|wx.LEFT|wx.TOP, border=4)
        topRightIconComboSizer.Add(self.remoteDirCombo, 1, wx.EXPAND|wx.LEFT, border=4)

        self.remoteFileList = FileListCtrl(topRightPanel, -1, style=wx.LC_REPORT)
#        self.remoteFileList.SetBackgroundColour(wx.Colour(200,240,200))

        topRightSizer = wx.BoxSizer(wx.VERTICAL)
        topRightSizer.Add(topRightIconComboSizer, 0, wx.EXPAND, 1)
        topRightSizer.Add(self.remoteFileList, 1, wx.EXPAND | wx.ALL, 1)
        topRightPanel.SetSizer(topRightSizer)

        self.vsplitter.SplitVertically(topLeftPanel, topRightPanel)
        

        # Create the bottom half of the horizontal splitter
        bottomPanel = wx.Panel(self.hsplitter, -1)
        bottomPanel.SetBackgroundColour('#E0E0E0')

        self.outputList = OutputList(bottomPanel, style=wx.LB_EXTENDED)

        bottomSizer = wx.BoxSizer(wx.HORIZONTAL)
        bottomSizer.Add(self.outputList, 1, wx.EXPAND | wx.ALL, 1)

        bottomPanel.SetSizer(bottomSizer)

        self.hsplitter.SplitHorizontally(self.vsplitter, bottomPanel)

        # Setup events
        self.Bind(wx.EVT_TEXT_ENTER, self.OnLocalDirText, id=ID_LOCAL_DIR_COMBO)
        self.Bind(wx.EVT_COMBOBOX, self.OnLocalDirText, id=ID_LOCAL_DIR_COMBO)
        self.Bind(wx.EVT_MENU, self.OnBrowseLocal, id=wx.ID_OPEN)
        self.Bind(wx.EVT_TEXT_ENTER, self.OnRemoteDirText, id=ID_REMOTE_DIR_COMBO)
        self.Bind(wx.EVT_COMBOBOX, self.OnRemoteDirText, id=ID_REMOTE_DIR_COMBO)
        self.Bind(wx.EVT_MENU, self.OnBrowseRemote, id=ID_REMOTE_CONNECT)
        self.Bind(wx.EVT_MENU, self.OnFileSettings, id=ID_FILE_SETTINGS)
        self.Bind(wx.EVT_MENU, self.OnOpenHelp, id=ID_OPEN_HELP)
        self.Bind(wx.EVT_MENU, self.OnOpenHelpAbout, id=ID_OPEN_HELP_ABOUT)
        self.Bind(wx.EVT_MENU, self.OnOpenHelpLicensing, id=ID_OPEN_HELP_LICENSING)

        self.Bind(wx.EVT_MENU, self.OnLocalUpload, id=ID_LOCAL_UPLOAD)
        self.Bind(wx.EVT_MENU, self.OnRemoteDownload, id=ID_REMOTE_DOWNLOAD)
        self.Bind(wx.EVT_SIZE, self.OnSize)
        
        # Open initial local directory
        homeDirPath = os.path.expanduser("~")
        localDirList = [homeDirPath]
        self.localFileList.Init(None, self.localDirCombo, self.outputList, self.remoteFileList, None)
        self.localFileList.SetDir(homeDirPath)
        self.localDirCombo.Append(homeDirPath)
        self.localDirCombo.SetStringSelection(homeDirPath)

        # Initialise the remote directory list control
        self.remoteFileList.Init(self.remoteIcon, self.remoteDirCombo, self.outputList, None, self.localFileList)

        self.SetSize((1200, 800))
        self.SetTitle('Cloud Filer')
        self.Centre()
#        self.Show(True)
        

#    def OnIdle(self, e):
#        global gbl_newerVersionAvailable
#        if gbl_newerVersionAvailable:
#            gbl_newerVersionAvailable = False
#            self.Unbind(wx.EVT_IDLE, handler=self.OnIdle)
#            dlg = Utils.InformationDlg(Utils.ACTION_NEW_VERSION, False, self, wx.ID_ANY, 'Cloud Filer - Information')
#            dlg.ShowModal()
#            if dlg.GetEnoughAlready():
#                AppConfig.Set('History', 'UserSupressedVersionCheck', 'Y')
#                AppConfig.WriteSettings()
#            dlg.Destroy()


    def OnSize(self, e):
        width, height = e.GetSize().Get()
        self.vsplitter.SetSashPosition(width / 2, True)
        self.hsplitter.SetSashPosition(height / 2, True)
        e.Skip()


    def OnLocalUpload(self, e):
        self.localFileList.OnUpload(None)


    def OnRemoteDownload(self, e):
        self.remoteFileList.OnDownload(None)


    def OnOpenHelp(self, e):
        DEBUG('OnOpenHelp: os.name=' + os.name)
        try:
            if GetOS() == ID_OS_WINDOWS:
                os.startfile(Utils.GetFilePath(USER_GUIDE_APPLICATION)) # Windows
            elif GetOS() == ID_OS_LINUX:
                os.system("xdg-open "+Utils.GetFilePath(USER_GUIDE_APPLICATION)) # Linux
            elif GetOS() == ID_OS_MACOS:
                os.system('open', filepath) # MacOS
            else:
                DEBUG('Unexpected OS')
        except:
            DEBUG('Open help exception (is a PDF reader installed?)')
            pass


    def OnOpenHelpLicensing(self, e):
        dlg = Utils.InformationDlg(Utils.ACTION_LICENSE_DETAILS, False, self, wx.ID_ANY, 'Cloud Filer - License Details')
        dlg.ShowModal()
        dlg.Destroy()


    def OnOpenHelpAbout(self, e):
        
        description = """
Cloud Filer is a cloud storage file management client.
Cloud Filer allows you to upload and download files to cloud
storage and can be used as part of a backup strategy.  For
information about updates check out the website.
"""

        licence = """
The MIT License (MIT)

Copyright (c) 2017 Nuvovis

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the
Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall
be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
        info = wx.AboutDialogInfo()

        info.SetName('Cloud Filer')
        info.SetVersion(CF_VERSION_STRING)
        info.SetDescription(description)
        info.SetCopyright('(C) 2017 Nuvovis')
        info.SetWebSite('http://www.nuvovis.com')
        info.SetLicence(licence)

        wx.AboutBox(info)


    def OnLocalDirText(self, e):
        newDir = self.localDirCombo.GetValue()
        if os.path.isdir(newDir):
            self.localFileList.SetDir(newDir)
        else:
            self.outputList.Append('Invalid dir: ' + newDir)


    def OnRemoteDirText(self, e):
        newDir = self.remoteDirCombo.GetValue()
        self.remoteFileList.SetRemoteCryptoDirFromCache(newDir)
        self.remoteFileList.ChangeRemoteDir(newDir)


    def OnBrowseLocal(self, e):
        dlg = wx.DirDialog(self, 'Select a local folder', style=wx.DD_CHANGE_DIR|wx.DD_DIR_MUST_EXIST)
        if dlg.ShowModal() == wx.ID_OK:
            self.localDirCombo.Append(dlg.GetPath())
            self.localDirCombo.SetStringSelection(dlg.GetPath())
            self.localFileList.SetDir(dlg.GetPath())

        dlg.Destroy()


    def CheckLicensed(self, proxyServer, proxyUsername, proxyPassword):
        if Licensing.CloudFilerLicensed():
            return True

        # Initialise licensing
        try:
            key = AppConfig.Get('History', 'LicenseKey')
            DEBUG('LicenseKey=' + key)
        except:
            key = ''
        Licensing.Init(os.path.realpath(Utils.get_main_dir()), key, proxyServer, proxyUsername, proxyPassword)
        if not Licensing.CloudFilerLicensed():
            dlg = GetNewLicenseDlg(Licensing.GetLicensingMsg(), self, wx.ID_ANY, 'License Cloud Filer')
            if dlg.ShowModal() == wx.ID_OK:
                DEBUG('Activating license...')
                key = dlg.GetKey()
                Licensing.Init(os.path.realpath(Utils.get_main_dir()), key, proxyServer, proxyUsername, proxyPassword)
                if Licensing.CloudFilerLicensed():
                    DEBUG('License retireved...')
                    AppConfig.Set('History', 'LicenseKey', key)
                    AppConfig.WriteSettings()
                    dlg2 = Utils.InformationDlg(Utils.ACTION_LICENSE_DETAILS, False, self, wx.ID_ANY, 'Cloud Filer - License Details')
                    dlg2.ShowModal()
                else:
                    DEBUG('License no good')
                    errorMsg = 'Failed to activate license: ' + Licensing.GetLicensingMsg()
                    errorMsg += '\nCheck proxy settings?'
                    dlg2 = wx.MessageDialog(None, errorMsg, 'Error', wx.OK | wx.ICON_ERROR)
                    dlg2.ShowModal()

            dlg.Destroy()

        return Licensing.CloudFilerLicensed()

    def OnBrowseRemote(self, e):
        if self.GetCloudCredentials():
            self.remoteFileList.S3Connect(self.accessKey, self.secretKey, self.password, self.proxyServer, self.proxyUsername, self.proxyPassword, self._cloudService, self._cloudHost, self._cloudRootText)
            if self.remoteFileList._connected:
                self.remoteFileList.ChangeRemoteDir(self._cloudRootText)


    def GetCloudCredentials(self):
        dlg = CloudConnectDlg(self, wx.ID_ANY, 'Specify cloud connection')
        if dlg.ShowModal() == wx.ID_OK:
            self.accessKey = dlg.GetAccessKey()
            self.secretKey = dlg.GetSecretKey()
            self.password = dlg.GetPassword()
            self._cloudService = dlg.GetCloudService()
            self._cloudHost = dlg.GetCloudHost()
            self._cloudRootText = dlg.GetRootText()
            keyKey = 'nuvovis.com CloudFiler ' + self._cloudRootText.rstrip(':')
            import unicodedata
            keyKey = unicodedata.normalize('NFKD', keyKey).encode('ascii', 'ignore')
            if dlg.StoreKeys():
                AppConfig.SecureSet(keyKey, 'AK', self.accessKey)
                AppConfig.SecureSet(keyKey, 'SK', self.secretKey)
            else:
                AppConfig.SecureSet(keyKey, 'AK', '')
                AppConfig.SecureSet(keyKey, 'SK', '')
            
            if dlg.StorePassword():
                AppConfig.SecureSet(keyKey, 'PWD', self.password)
            else:
                AppConfig.SecureSet(keyKey, 'PWD', '')

            if not dlg.GetEncrypt():
                self.password = '' # Switch off encryption now that we have saved the password if necessary

            dlg.SaveLastConnection() 
# With later keyring can use delete_password
# keyring.delete_password('AWSS3', 'AK')
# keyring.delete_password('AWSS3', 'SK')

#            self.remoteFileList.SetRemoteDir(self._cloudRootText)
            hitOk = True

        else:
            hitOk = False

        dlg.Destroy()
        return hitOk


    def OnFileSettings(self, e):
        dlg = SettingsDlg(self, wx.ID_ANY, 'Settings')
        if dlg.ShowModal() == wx.ID_OK:
            self.proxyServer = dlg.GetProxyServer()
            self.proxyUsername = dlg.GetProxyUsername()
            self.proxyPassword = dlg.GetProxyPassword()
            AppConfig.Set('History', 'ProxyServer', self.proxyServer)
            AppConfig.Set('History', 'ProxyUsername', self.proxyUsername)
            AppConfig.Set('History', 'ProxyPassword', self.proxyPassword)
            AppConfig.WriteSettings()

        dlg.Destroy()


    def OnQuit(self, e):
        Licensing.UnInit(os.path.realpath(Utils.get_main_dir()))
        self.Close()



def main():
    app = wx.App(False)  # Create a new app, don't redirect stdout/stderr to a window.
    locale = wx.Locale(wx.LANGUAGE_ENGLISH)
    frame = MainForm(None)
    frame.Show(True)     # Show the frame.
    app.MainLoop()


if __name__ == '__main__':
    main()

