import sys
import os
import wx
import Utils
import time
import threading
import tempfile

from boto.s3.connection import S3Connection
from boto.s3.key import Key
from boto.exception import S3ResponseError
from boto.exception import S3CreateError
from base64 import b64decode
from base64 import b64encode
from passlib.hash import pbkdf2_sha256
from FileCrypto import EncryptStr
from FileCrypto import DecryptStr
from FileCrypto import EncryptFile
from FileCrypto import DecryptFile
from ast import literal_eval

CTA_ALTCHARS = './'
FOLDER_CONTENTS_HEADER = 'Nuvovis Cloud Filer folder contents:'
SORT_BY_FILENAME = 0
SORT_BY_FILESIZE = 1
SORT_BY_FILETYPE = 2
SORT_BY_MODIFIED = 3

def DEBUG(msg):
# Uncomment the following line for debug output
    print msg.encode('utf-8', errors='replace')
#    pass

ID_EVT_ACTION = 108
def EVT_ACTION(win, func):
    """Define Action Event."""
    win.Connect(-1, -1, ID_EVT_ACTION, func)


class ActionEvent(wx.PyEvent):
    """Simple event to carry arbitrary action data."""
    def __init__(self, data):
        wx.PyEvent.__init__(self)
        self.SetEventType(ID_EVT_ACTION)
        self.data = data

def percent_cb(complete, total):
    global gauge
    gauge.SetValue(complete)

def filename_from_path(path):
    lastSlashPos = path.rfind('/')
    if lastSlashPos >= 0:
        return path[lastSlashPos+1:]
    else:
        return path



class CloudActionDlg(wx.Dialog):

    def __init__(self, action, *args, **kwargs):
        super(CloudActionDlg, self).__init__(*args, **kwargs) 
        
        self.action = action
        self.InitUI()

        
    def InitUI(self):
        self.SetSize((400, 160))

        self.label1 = wx.StaticText(self, pos=(10, 15))

        global gauge
        gauge = wx.Gauge(self, size=(380, 20), pos=(10, 45))
        self.cloudActionCancelled = False

        cancelButton = wx.Button(self, id=wx.ID_CANCEL, label='Cancel', pos=(300, 90))
        cancelButton.SetFocus()

        cancelButton.Bind(wx.EVT_BUTTON, self.OnCancel)
        cancelButton.Bind(wx.EVT_BUTTON, self.OnCancel)
        
        EVT_ACTION(self, self.OnAction)


    def OnAction(self, e):
        self.SetActionMsg(e.data)


    def SetActionMsg(self, objectName):
        if len(objectName) > 56:
            objectName = objectName[:56] + '...'

        if self.action == Utils.ACTION_DELETE_BUCKET:
            msg = 'Deleting ' + objectName
        elif self.action == Utils.ACTION_UPLOAD_FILE:
            msg = 'Uploading ' + objectName
        elif self.action == Utils.ACTION_DOWNLOAD_FILE:
            msg = 'Downloading ' + objectName
        else:
            msg = 'Handling ' + objectName

        self.label1.SetLabel(msg)

        
    def OnCancel(self, e):        
        endId = e.GetId()
        if endId == wx.ID_CANCEL:
            DEBUG('cancel')
            self.cloudActionCancelled = True
            self.label1.SetLabel('Cancelling...')

            if self.action == Utils.ACTION_UPLOAD_FILE or self.action == Utils.ACTION_DOWNLOAD_FILE:
                done = False
                count = 100
                while not done and count > 0:
                    count = count - 1
                    global gbl_cloud_fp
                    try:
                        DEBUG('try closing fp')
                        gbl_cloud_fp.close()
                        done = True
                    except:
                        DEBUG('bombed')

        elif endId == wx.ID_CANCEL:
            self.Destroy()


    def GetCancelled(self):
        return self.cloudActionCancelled


class UploadThread (threading.Thread):

    def __init__(self, dlg, fileList):
        threading.Thread.__init__(self)
        self.dlg = dlg
        self.fileList = fileList
        self.bucket, self.keyName = self.fileList._remoteDirList.GetCurrentRemoteLocation()
        self.messages = unicode()


    def GetSomethingUploaded(self):
        return self.somethingUploaded


    def GetMessages(self):
        return self.messages


    def run(self):
        DEBUG("Waiting for dialog")
        while not self.dlg.IsModal():
            time.sleep(0.1)

        DEBUG("Starting upload thread")
        self.inError = False
        cancelled = False
        currentItem = self.fileList.GetFirstSelected()
        while currentItem != -1 and not cancelled:
            fileName = self.fileList.GetItem(currentItem, 0).GetText()
            fileType = self.fileList.GetItem(currentItem, 2).GetText()
            DEBUG('fileName=' + fileName)
            DEBUG('fileType=' + fileType)
            if fileType == 'Folder':
                DEBUG('it is a folder')
                cancelled, self.somethingUploaded = self.UploadFolder(self.bucket, self.keyName, fileName, self.dlg)
            else:
                DEBUG('it is a file')
                cancelled, self.somethingUploaded = self.UploadFile(self.bucket, self.keyName, fileName, self.dlg)

            currentItem = self.fileList.GetNextSelected(currentItem)
        # Close the upload dialog
        wx.PostEvent(self.dlg, wx.PyCommandEvent(wx.EVT_BUTTON.typeId, wx.ID_CANCEL))


    def UploadFile(self, bucket, keyName, fileName, dlg):
        if self.dlg.GetCancelled():
            self.messages = self.messages + 'Upload cancelled' + '\n'
            return True, False

        DEBUG('UploadFile: fileName=' + fileName + '; keyName=' + keyName)
        cancelled = False
        somethingUploaded = False

        justFileName = filename_from_path(fileName)
        if justFileName.startswith('_CF'):
            self.inError = True
            self.errorMessage = 'Skipping file that looks like a CloudFiler encrypted file: ' + fileName
            DEBUG(self.errorMessage)
        else:
            tempFilePath = ''
            try:
                wx.PostEvent(self.dlg, ActionEvent(fileName))
                fileSize = os.path.getsize(fileName)
                DEBUG('fileSize=' + str(fileSize))
                global gauge
                gauge.SetRange(fileSize)

                global gbl_cloud_fp
                if len(self.fileList._remoteDirList._password) == 0:
                    gbl_cloud_fp = open(fileName, 'rb')
                else:
                    DEBUG('encrypt to temp file')
                    sec = os.path.getmtime(fileName)
                    timeStr = time.strftime('%Y-%m-%d %H:%M', time.localtime(sec))
                    gbl_cloud_fp = open(fileName, 'rb')
                    cryptoFileName = self.fileList._remoteDirList.GetAndRemoveExistingCryptoFileName(fileName)
                    if cryptoFileName is None:
                        cryptoFileName = self.fileList._remoteDirList.GetNewCryptoFileName()
                    tempFilePath = tempfile.gettempdir() + os.sep + cryptoFileName
                    tmpfp = open(tempFilePath, 'wb')
                    EncryptFile(gbl_cloud_fp, tmpfp, self.fileList._remoteDirList._encryptionKey)
                    gbl_cloud_fp.close()
                    tmpfp.close()
                    plaintextFileName = fileName
                    fileName = cryptoFileName
                    gbl_cloud_fp = open(tempFilePath, 'rb')

                DEBUG('bucket name=' + self.bucket.name)
                DEBUG('key name=' + keyName)
                DEBUG('pos=' + str(gbl_cloud_fp.tell()))
                k = Key(self.bucket)
                k.key = keyName + fileName.replace('\\', '/')
                DEBUG('key=' + k.key)
                k.set_contents_from_file(gbl_cloud_fp, cb=percent_cb, num_cb=fileSize)
                gbl_cloud_fp.close()

                if len(self.fileList._remoteDirList._password) > 0:
                    self.fileList._remoteDirList.AddedCryptoFile(cryptoFileName, plaintextFileName, str(fileSize), timeStr)
                    self.fileList._remoteDirList.UploadNewContentsInfo()

            except S3ResponseError as ex:
                self.inError = True
                if ex.error_message != None and len(ex.error_message) > 0:
                    msg = ex.error_message
                else:
                    msg = str(ex)
                self.errorMessage = 'S3 Error (cloud action): "' + msg + '"'
                DEBUG('S3 Cloud thread error: ' + self.errorMessage)

            except Exception as ex:
                self.inError = True
                self.errorMessage = 'Error (cloud action): ' + str(ex)
                DEBUG('Cloud thread error: ' + self.errorMessage)

            finally:
                if len(tempFilePath) > 0:
                    try:
                        os.unlink(tempFilePath)
                    except Exception as ex:
                        DEBUG('Unlink temp file error: ' + str(ex))

        if self.dlg.GetCancelled():
            self.messages = self.messages + 'Upload cancelled' + '\n'
            cancelled = True
        else:
            if self.inError:
                self.messages = self.messages + 'Upload of ' + fileName + ' failed - ' + self.errorMessage + '\n'
            else:
                if len(self.fileList._remoteDirList._password) > 0:
                    fileName = self.fileList._remoteDirList.GetCleartextFileName(fileName)
                self.messages = self.messages + 'Uploaded ' + fileName + '\n'
                somethingUploaded = True

        return cancelled, somethingUploaded



    def UploadFolder(self, bucket, keyName, folderName, dlg):
        if self.dlg.GetCancelled():
            self.messages = self.messages + 'Upload cancelled' + '\n'
            return True, False

        if not keyName.startswith('/'):
            keyName = '/' + keyName
        DEBUG('UploadFolder: folderName=' + folderName + '; keyName=' + keyName)

        lastSlashPos = folderName.rfind('/')
        if lastSlashPos >= 0:
            justfolderName = folderName[lastSlashPos+1:]
        else:
            justfolderName = folderName
        if justfolderName.startswith('_CF'):
            cancelled = False
            somethingUploaded = False
            self.messages = self.messages + 'Skipping folder that looks like a CloudFiler encrypted folder: ' + folderName + '\n'
        else:
            self.messages = self.messages + 'Uploading ' + folderName + ' to ' + bucket.name + keyName + '\n'

            originalLocalDir = ''
            originalRemoteDir = ''
            originalRemoteCryptoDir = ''
            # If we are in encrypted mode 
            if len(self.fileList._remoteDirList._password) > 0:

                # if there is a '/' char in the folderName
                firstSlashPos = folderName.find('/')
                assert (firstSlashPos != 0), "folderName started with slash"
                if firstSlashPos > 0:
                    # Get top folder name and remove top folder name from folderName
                    topFolderName = folderName[:firstSlashPos]
                    DEBUG('UploadFolder: topFolderName=' + topFolderName)
                    folderName = folderName[firstSlashPos+1:]
                    DEBUG('UploadFolder: folderName=' + folderName)
                else:
                    DEBUG('folderName is topFolderName')
                    topFolderName = folderName
                    DEBUG('UploadFolder: topFolderName=' + topFolderName)
                    folderName = '.'
                    DEBUG('UploadFolder: folderName=' + folderName)

                # If the remote top folder name does not exist i.e. is not in cryptoDirs
                if not self.fileList._remoteDirList.CryptoDirNameExists(topFolderName):
                    DEBUG('UploadFolder: create new remote crypto folder')
                    cryptoFolderName = self.fileList._remoteDirList.GetNewCryptoFolderName()
                    self.fileList._remoteDirList.CreateNewFolder(cryptoFolderName)
                    self.fileList._remoteDirList.AddedCryptoDir(cryptoFolderName, topFolderName)
                    self.fileList._remoteDirList.UploadNewContentsInfo()
                else:
                    DEBUG('UploadFolder: get existing remote crypto folder')
                    cryptoFolderName = self.fileList._remoteDirList.GetCryptoFolder(topFolderName)

                # move position to remote folder called folderName
                originalRemoteDir = self.fileList._remoteDirList._currentRemotePath
                originalRemoteCryptoDir = self.fileList._remoteDirList._currentRemoteCryptoPath
                self.fileList._remoteDirList._currentRemotePath = originalRemoteDir + '/' + topFolderName
                self.fileList._remoteDirList._currentRemoteCryptoPath = originalRemoteCryptoDir + '/' + cryptoFolderName
                DEBUG('UploadFolder: cd currentRemotePath=' + self.fileList._remoteDirList._currentRemotePath)
                DEBUG('UploadFolder: cd currentRemoteCryptoPath=' + self.fileList._remoteDirList._currentRemoteCryptoPath)
                bucket, keyName = self.fileList._remoteDirList.GetCurrentRemoteLocation()
                self.fileList._remoteDirList.GetEncryptedDirContents(bucket, keyName)

                # move local position to local folder called folderName
                originalLocalDir = os.getcwd()
                dirPath = originalLocalDir + '/' + topFolderName
                DEBUG('UploadFolder: lcd dirPath=' + dirPath)
                os.chdir(dirPath)

            cancelled = False
            somethingUploaded = False
            # Skip any folder we can't access
            try:
                files = sorted(os.listdir(Utils.ensure_unicode(folderName)))
                for i in files:
                    DEBUG('UploadFolder: i=' + i)
                    if folderName == '.':
                        filePath = i
                    else:
                        filePath = folderName + os.sep + i
                    if os.path.isdir(filePath):
                        cancelled, somethingUploaded = self.UploadFolder(bucket, keyName, filePath, dlg)
                    else:
                        cancelled, somethingUploaded = self.UploadFile(bucket, keyName, filePath, dlg)
                    if cancelled:
                        break

            except Exception as ex:
                self.messages = self.messages + 'Skipping folder that I failed to read\n'

            # Pop directory location back to parent
            if len(originalLocalDir) > 0:
                os.chdir(originalLocalDir)
            if len(originalRemoteDir) > 0:
                self.fileList._remoteDirList._currentRemotePath = originalRemoteDir
            if len(originalRemoteCryptoDir) > 0:
                self.fileList._remoteDirList._currentRemoteCryptoPath = originalRemoteCryptoDir

            # Update the crypto directory contents lists
            bucket, keyName = self.fileList._remoteDirList.GetCurrentRemoteLocation()
            # If we are in encrypted mode 
            if len(self.fileList._remoteDirList._password) > 0:
                self.fileList._remoteDirList.GetEncryptedDirContents(bucket, keyName)

        return cancelled, somethingUploaded



class DownloadThread (threading.Thread):

    def __init__(self, dlg, fileList):
        threading.Thread.__init__(self)
        self.dlg = dlg
        self.fileList = fileList
        self.bucket, self.keyName = self.fileList.GetCurrentRemoteLocation()
        if self.keyName.startswith('/'):
            self.keyName = self.keyName.lstrip('/')
        if len(self.keyName) > 0 and not self.keyName.endswith('/'):
            self.keyName = self.keyName + '/'
        self.messages = ''


    def GetSomethingDownloaded(self):
        return self.somethingDownloaded


    def GetMessages(self):
        return self.messages


    def run(self):
        DEBUG("Waiting for dialog")
        while not self.dlg.IsModal():
            time.sleep(0.1)

        DEBUG("Starting download thread")
        self.inError = False
        cancelled = False
        currentItem = self.fileList.GetFirstSelected()
        while currentItem != -1 and not cancelled:
            fileName = self.fileList.GetItem(currentItem, 0).GetText()
            fileType = self.fileList.GetItem(currentItem, 2).GetText()
            DEBUG('fileName=' + fileName)
            DEBUG('fileType=' + fileType)
            if fileType == 'Folder':
                DEBUG('it is a folder')
                cancelled, self.somethingDownloaded = self.DownloadFolder(self.bucket, self.keyName, fileName, self.dlg)
            else:
                DEBUG('it is a file')
                cancelled, self.somethingDownloaded = self.DownloadFile(self.bucket, self.keyName, fileName, self.dlg)

            currentItem = self.fileList.GetNextSelected(currentItem)
        # Close the upload dialog
        wx.PostEvent(self.dlg, wx.PyCommandEvent(wx.EVT_BUTTON.typeId, wx.ID_CANCEL))


    def DownloadFile(self, bucket, keyName, fileName, dlg):
        DEBUG('DownloadFile: keyName = ' + keyName)
        DEBUG('DownloadFile: fileName = ' + fileName)

        try:
            cancelled = False
            somethingDownloaded = False
            # If not encrypted mode
            if len(self.fileList._password) == 0 or fileName.startswith('_CF'):
                downloadFileName = fileName
                decrypting = False
            else:
                downloadFileName = self.fileList.GetCryptoFileName(fileName)
                decrypting = True
            fileKey = keyName + downloadFileName.replace('\\', '/')
            DEBUG('DownloadFile: fileKey = ' + fileKey)
            k = self.bucket.get_key(fileKey)
            DEBUG('key=' + k.key)
            wx.PostEvent(self.dlg, ActionEvent(fileName))
            fileSize = k.size
            DEBUG('fileSize=' + str(fileSize))
            
            global gauge
            gauge.SetRange(fileSize)

            DEBUG('bucket name=' + self.bucket.name)
            DEBUG('key name=' + keyName)
            global gbl_cloud_fp
            gbl_cloud_fp = open(downloadFileName, 'wb')
            k.get_contents_to_file(gbl_cloud_fp, cb=percent_cb, num_cb=fileSize)
            gbl_cloud_fp.close()
            if decrypting:
                tmpfp = open(downloadFileName, 'rb') # Open the local obfuscated _CFLn file we have just downloaded
                fp = open(fileName, 'wb') # Open the file we really want to write to
                DecryptFile(tmpfp, fp, self.fileList._encryptionKey)
                fp.close()
                tmpfp.close()
                os.unlink(downloadFileName)

        except S3ResponseError as ex:
            self.inError = True
            if ex.error_message != None and len(ex.error_message) > 0:
                msg = ex.error_message
            else:
                msg = str(ex)
            self.errorMessage = 'S3 Error (cloud action): "' + msg + '"'
            DEBUG('S3 Cloud thread error: ' + self.errorMessage)

        except Exception as ex:
            self.inError = True
            self.errorMessage = 'Error (cloud action): ' + str(ex)
            DEBUG('Cloud thread error: ' + self.errorMessage)

        if self.dlg.GetCancelled():
            self.messages = self.messages + 'Download cancelled' + '\n'
            cancelled = True
        else:
            if self.inError:
                self.messages = self.messages + 'Download of ' + fileName + ' failed - ' + self.errorMessage + '\n'
            else:
                self.messages = self.messages + 'Downloaded ' + fileName + '\n'
                somethingDownloaded = True

        return cancelled, somethingDownloaded



    def DownloadFolder(self, bucket, keyName, folderName, dlg):
        DEBUG('DownloadFolder: keyName=' + keyName)
        DEBUG('DownloadFolder: folderName=' + folderName)
        self.messages = self.messages + 'Downloading ' + folderName + ' from ' + bucket.name + '/' + keyName + '\n'
        try:
            os.mkdir(folderName)

        except OSError as ex:
            # If this is an exception because the folder or file already exists
            if 'exists' in str(ex):
                self.messages = self.messages + 'Merging into existing folder: ' + folderName + '\n'

        # If we are in encrypted mode 
        if len(self.fileList._password) > 0:
            return self.DownloadFolderInCryptoMode(bucket, keyName, folderName, dlg)
        else:
            return self.DownloadFolderInCleartextMode(bucket, keyName, folderName, dlg)



    def DownloadFolderInCleartextMode(self, bucket, keyName, folderName, dlg):
        originalLocalDir = ''
        originalRemoteDir = ''

        # if there is a '/' char in the folderName
        firstSlashPos = folderName.find('/')
        assert (firstSlashPos != 0), "folderName started with slash"
        if firstSlashPos > 0:
            # Get top folder name and remove top folder name from folderName
            topFolderName = folderName[:firstSlashPos]
            DEBUG('DownloadFolder: topFolderName=' + topFolderName)
            folderName = folderName[firstSlashPos+1:]
            DEBUG('DownloadFolder: folderName=' + folderName)
        else:
            DEBUG('folderName is topFolderName')
            topFolderName = folderName
            DEBUG('DownloadFolder: topFolderName=' + topFolderName)
            folderName = '.'
            DEBUG('DownloadFolder: folderName=' + folderName)

        # move position to remote folder called folderName
        originalRemoteDir = self.fileList._currentRemotePath
        self.fileList._currentRemotePath = originalRemoteDir + '/' + topFolderName
        DEBUG('DownloadFolder: cd currentRemotePath=' + self.fileList._currentRemotePath)
        bucket, keyName = self.fileList.GetCurrentRemoteLocation()

        # move local position to local folder called folderName
        originalLocalDir = os.getcwd()
        dirPath = originalLocalDir + '/' + topFolderName
        DEBUG('DownloadFolder: lcd dirPath=' + dirPath)
        os.chdir(dirPath)

        cancelled = False
        somethingDownloaded = False

        DEBUG('listing files for download at: ' + keyName)
        files = bucket.list(keyName, '/')
        for fileEntry in files:
            if fileEntry.name == keyName + '.$folder$' or fileEntry.name == keyName:
                continue
            DEBUG('file or folder for download: ' + fileEntry.name)
            # If it's a folder
            if fileEntry.name.endswith('/'):
                justFolderName = filename_from_path(fileEntry.name.rstrip('/'))
                DEBUG('Folder for download: ' + justFolderName)
                cancelled, somethingDownloaded = self.DownloadFolder(bucket, keyName, justFolderName, dlg)
            else:
                justFileName = filename_from_path(fileEntry.name)
                DEBUG('File for download: ' + justFileName)
                cancelled, somethingDownloaded = self.DownloadFile(bucket, keyName, justFileName, dlg)
            if cancelled:
                break

        # Pop directory location back to parent
        if len(originalLocalDir) > 0:
            os.chdir(originalLocalDir)
        if len(originalRemoteDir) > 0:
            self.fileList._currentRemotePath = originalRemoteDir

        return cancelled, somethingDownloaded



    def DownloadFolderInCryptoMode(self, bucket, keyName, folderName, dlg):
        originalLocalDir = ''
        originalRemoteDir = ''
        originalRemoteCryptoDir = ''

        # if there is a '/' char in the folderName
        firstSlashPos = folderName.find('/')
        assert (firstSlashPos != 0), "folderName started with slash"
        if firstSlashPos > 0:
            # Get top folder name and remove top folder name from folderName
            topFolderName = folderName[:firstSlashPos]
            DEBUG('DownloadFolder: topFolderName=' + topFolderName)
            folderName = folderName[firstSlashPos+1:]
            DEBUG('DownloadFolder: folderName=' + folderName)
        else:
            DEBUG('folderName is topFolderName')
            topFolderName = folderName
            DEBUG('DownloadFolder: topFolderName=' + topFolderName)
            folderName = '.'
            DEBUG('DownloadFolder: folderName=' + folderName)

        # If the remote top folder name does not exist i.e. is not in cryptoDirs
        if not self.fileList.CryptoDirNameExists(topFolderName):
            raise Exception('CryptoDirNameExists is false for ' + topFolderName)
        else:
            DEBUG('DownloadFolder: get existing remote crypto folder')
            cryptoFolderName = self.fileList.GetCryptoFolder(topFolderName)

        # move position to remote folder called folderName
        originalRemoteDir = self.fileList._currentRemotePath
        originalRemoteCryptoDir = self.fileList._currentRemoteCryptoPath
        self.fileList._currentRemotePath = originalRemoteDir + '/' + topFolderName
        self.fileList._currentRemoteCryptoPath = originalRemoteCryptoDir + '/' + cryptoFolderName
        DEBUG('DownloadFolder: cd currentRemotePath=' + self.fileList._currentRemotePath)
        DEBUG('DownloadFolder: cd currentRemoteCryptoPath=' + self.fileList._currentRemoteCryptoPath)
        bucket, keyName = self.fileList.GetCurrentRemoteLocation()
        self.fileList.GetEncryptedDirContents(bucket, keyName)

        # move local position to local folder called folderName
        originalLocalDir = os.getcwd()
        dirPath = originalLocalDir + '/' + topFolderName
        DEBUG('DownloadFolder: lcd dirPath=' + dirPath)
        os.chdir(dirPath)

        cancelled = False
        somethingDownloaded = False
        for (s1, s2, s3, s4) in self.fileList.cryptoFiles:
            DEBUG('File for download: ' + s2)
            cancelled, somethingDownloaded = self.DownloadFile(bucket, keyName, s2, dlg)
            if cancelled:
                break

        for (s1, s2, s3, s4) in self.fileList.cryptoDirs:
            DEBUG('Folder for download: ' + s2)
            cancelled, somethingDownloaded = self.DownloadFolder(bucket, keyName, s2, dlg)
            if cancelled:
                break

        # Pop directory location back to parent
        if len(originalLocalDir) > 0:
            os.chdir(originalLocalDir)
        if len(originalRemoteDir) > 0:
            self.fileList._currentRemotePath = originalRemoteDir
        if len(originalRemoteCryptoDir) > 0:
            self.fileList._currentRemoteCryptoPath = originalRemoteCryptoDir

        # Update the crypto directory contents lists
        bucket, keyName = self.fileList.GetCurrentRemoteLocation()
        # If we are in encrypted mode 
        if len(self.fileList._password) > 0:
            self.fileList.GetEncryptedDirContents(bucket, keyName)

        return cancelled, somethingDownloaded


class DeleteThread (threading.Thread):

    def __init__(self, dlg, keys):
        threading.Thread.__init__(self)
        self.dlg = dlg
        self.keys = keys
        self.messages = ''


    def run(self):
        DEBUG("Waiting for dialog")
        while not self.dlg.IsModal():
            time.sleep(0.1)

        DEBUG("Starting delete thread")

        for key in self.keys:
            DEBUG('key.name=' + key.name)
            if key.name.endswith('/'):
                DEBUG('Ignoring dir')
                continue
            self.DeleteRemoteFile(key)
            global gauge
            gauge.Pulse()
            if self.dlg.GetCancelled():
                break

        # Close the upload dialog
        wx.PostEvent(self.dlg, wx.PyCommandEvent(wx.EVT_BUTTON.typeId, wx.ID_CANCEL))
        

    def GetMessages(self):
        return self.messages


    def DeleteRemoteFile(self, key):
        DEBUG('DeleteRemoteFile: key=' + key.name)
        try:
            wx.PostEvent(self.dlg, ActionEvent(key.name))
            self.messages = self.messages + 'Deleted file ' + key.name + '\n'
            # Delete key
            key.delete()

        except S3ResponseError as ex:
            errorMessage = ex.error_message
            self.messages = self.messages + 'S3 Error: "' + errorMessage + '"' + '\n'

        except Exception as ex:
            errorMessage = str(ex)
            self.messages = self.messages + 'Error: ' + errorMessage + '\n'



class FileListCtrl(wx.ListCtrl):

    def __init__(self, *args, **kwargs):
        super(FileListCtrl, self).__init__(*args, **kwargs) 
            
        self._connected = False
        self._currentRemotePath = ''
        self._currentRemoteCryptoPath = ''
        self._cryptoDirMap = {}

        self.InitUI()
     

    def InitUI(self):

        self.InsertColumn(0, 'Name', width = 270)
        self.InsertColumn(1, 'Size', width = 100)
        self.InsertColumn(2, 'Type', width = 100)
        self.InsertColumn(3, 'Modified', width = 120)

        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.DoubleClick)

        self.il = wx.ImageList(16, 16)
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_UP_DIR)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_FOLDER)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_PLAIN)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_LOCK)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_TEXT)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_PDF)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_IMAGE)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_PACKAGE)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_MUSIC)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_FILM)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_SUBTITLES)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_CDROM)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_SCRIPT_PYTHON)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_WORD)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_EXCEL)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_POWERPOINT)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DOC_HTML)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_SIGNATURE)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_REMOTE_FOLDER)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_CSS)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_SCRIPT)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_EXE)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_LIB)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_DLL)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_SMALL_ARROW_DOWN)))
        self.il.Add(wx.Bitmap(Utils.GetFilePath(Utils.ICON_SMALL_ARROW_UP)))

        self.SetImageList(self.il, wx.IMAGE_LIST_SMALL)

        self._sortAscending = True
        self.sortColumn = SORT_BY_FILENAME
        self.SwitchOnSortIndicator()

        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)
        self.Bind(wx.EVT_LIST_COL_CLICK, self.OnColumnClick)


    def Init(self, remoteIcon, comboBox, outputList, rdl, ldl):
        self._remoteIcon = remoteIcon
        self._combo = comboBox
        self._outputList = outputList
        self._remoteDirList = rdl
        self._localDirList = ldl
        self._connected = False
        self.buckets = None
        self._encryptionKey = ''
        self._salt = ''


    def OnViewRefresh(self, e):
        # If this is the remote list
        if self._remoteDirList == None:
            self.RefreshRemoteDir()
        else:
            self.RefreshDir()


    def SwitchOffSortIndicator(self):
        self.ClearColumnImage(self.sortColumn)


    def SwitchOnSortIndicator(self):
        if self._sortAscending:
            self.SetColumnImage(self.sortColumn, Utils.ICON_SMALL_ARROW_DOWN)
        else:
            self.SetColumnImage(self.sortColumn, Utils.ICON_SMALL_ARROW_UP)


    def ResetColumnSorting(self):
        self.SwitchOffSortIndicator()
        self._sortAscending = True
        self.sortColumn = SORT_BY_FILENAME
        self.SwitchOnSortIndicator()


    def OnColumnClick(self, e):
        DEBUG('OnColumnClick: ' + str(e.GetColumn()))

        self.DeleteAllItems()
        # if user clicked on the current column
        if e.GetColumn() == self.sortColumn:
            # Toggle sort order and change title
            if self._sortAscending:
                self._sortAscending = False
            else:
                self._sortAscending = True
            self.SwitchOnSortIndicator()

        else:
            self.SwitchOffSortIndicator()
            # Switch back to normal ascending sort
            self._sortAscending = True
            # Set new sort column
            self.sortColumn = e.GetColumn()
            self.SwitchOnSortIndicator()

        self.DisplayFileList()



    def OnRightDown(self, e):
        DEBUG('OnRightDown')
        try:
            # If have selected a single folder or bucklet then add Open action
            fileName = ''
            self.menu = wx.Menu()
            mi = wx.MenuItem(self.menu, wx.NewId(), 'Refresh')
            self.menu.AppendItem(mi)
            self.Bind(wx.EVT_MENU, self.OnViewRefresh, mi)
            if self.GetSelectedItemCount() == 1:
                fileName = self.GetItem(self.GetFirstSelected(), 0).GetText()
                fileType = self.GetItem(self.GetFirstSelected(), 2).GetText()
                DEBUG('fileName=' + fileName)
                DEBUG('fileType=' + fileType)
                if (fileType == 'Folder' or fileType == 'Bucket'):
                    mi = wx.MenuItem(self.menu, wx.NewId(), 'Open')
                    self.menu.AppendItem(mi)
                    self.Bind(wx.EVT_MENU, self.DoubleClick, mi)

            # If this is the remote list
            if self._remoteDirList == None:
            
                # If we are connected
                if self._connected:
                    pathBits = self._currentRemotePath.split('/')

                    if self.menu.GetMenuItemCount() > 0:
                        self.menu.AppendSeparator()

                    # If at root showing bucket list
                    if len(pathBits) < 2 or pathBits[1] == '':
                        mi = wx.MenuItem(self.menu, wx.NewId(), 'New Bucket')
                        self.menu.AppendItem(mi)
                        self.Bind(wx.EVT_MENU, self.OnNewBucket, mi)
                        if self.GetSelectedItemCount() > 0:
                            mi = wx.MenuItem(self.menu, wx.NewId(), 'Delete')
                            self.menu.AppendItem(mi)
                            self.Bind(wx.EVT_MENU, self.OnDeleteBucket, mi)
                    else:

                        if self.GetSelectedItemCount() == 1:
                            fileType = self.GetItem(self.GetFirstSelected(), 2).GetText()
                            DEBUG('fileType=' + fileType)

                        mi = wx.MenuItem(self.menu, wx.NewId(), 'New Folder')
                        self.menu.AppendItem(mi)
                        self.Bind(wx.EVT_MENU, self.OnNewFolder, mi)
                        if self.GetSelectedItemCount() > 0:
                            mi = wx.MenuItem(self.menu, wx.NewId(), 'Delete')
                            self.menu.AppendItem(mi)
                            self.Bind(wx.EVT_MENU, self.OnDelete, mi)
                            self.menu.AppendSeparator()
                            mi = wx.MenuItem(self.menu, wx.NewId(), 'Download')
                            self.menu.AppendItem(mi)
                            self.Bind(wx.EVT_MENU, self.OnDownload, mi)

            # Else this is the local list
            else:
                # If something is selected
                if self.GetSelectedItemCount() > 0:
                    # If the remote side can accept stuff and the up dir (always first) is not selected
                    if self._remoteDirList.CanUpload() and self.GetItemText(self.GetFirstSelected()) != '..':
                        if self.menu.GetMenuItemCount() > 0:
                            self.menu.AppendSeparator()
                        mi = wx.MenuItem(self.menu, wx.NewId(), 'Upload')
                        self.menu.AppendItem(mi)
                        self.Bind(wx.EVT_MENU, self.OnUpload, mi)

            if self.menu.GetMenuItemCount() > 0:
                self.PopupMenu(self.menu, e.GetPosition())

        except Exception as ex:
            errorMessage = str(ex)
            self._outputList.Append('Error: ' + errorMessage)


    def CanUpload(self):
        if self._connected:
            # If not at bucket list
            pathBits = self._currentRemotePath.split('/')
            if len(pathBits) > 1 and pathBits[1] != '':
                return True

        return False


    def OnUpload(self, e):
        dlg = CloudActionDlg(Utils.ACTION_UPLOAD_FILE, self, wx.ID_ANY, 'Cloud Filer - Upload')
        thread = UploadThread(dlg, self)
        thread.start()
        dlg.ShowModal()
        thread.join()
    
        self._outputList.Append(thread.GetMessages())

        self._remoteDirList.RefreshRemoteDir()
        dlg.Destroy()
            

    def OnDownload(self, e):
        dlg = CloudActionDlg(Utils.ACTION_DOWNLOAD_FILE, self, wx.ID_ANY, 'Cloud Filer - Download')
        thread = DownloadThread(dlg, self)
        thread.start()
        dlg.ShowModal()
        thread.join()
    
        self._outputList.Append(thread.GetMessages())

        self._localDirList.RefreshDir()
        dlg.Destroy()



    def OnNewBucket(self, e):
        dlg = NewBucketOrFolderDlg(True, self, wx.ID_ANY, 'Create new bucket')
        if dlg.ShowModal() == wx.ID_OK:
            name = dlg.GetName()
            try:
                self._connection.create_bucket(name)
                self.buckets = self._connection.get_all_buckets()
                self.ChangeRemoteDir(self._cloudRootText)

            except S3CreateError as ex:
                self._outputList.Append('S3 Create Error: bucket name must be globally unique!')

            except S3ResponseError as ex:
                if len(ex.error_message) > 0:
                    errorMessage = ex.error_message
                else:
                    errorMessage = str(ex)
                self._outputList.Append('S3 Error: "' + errorMessage + '"')

            except Exception as ex:
                errorMessage = str(ex)
                self._outputList.Append('Error: ' + errorMessage)

        dlg.Destroy()


    def CreateNewFile(self, fileName, contents):
        try:
            bucket, keyName = self.GetCurrentRemoteLocation()
            if not keyName.startswith('/'):
                keyName = '/' + keyName
            keyName = keyName + '/'
            k = Key(bucket)
            k.key = keyName + fileName.replace('\\', '/')
            DEBUG('key=' + k.key)
            DEBUG('contents=' + contents)
            k.set_contents_from_string(contents)

        except S3ResponseError as ex:
            if len(ex.error_message) > 0:
                errorMessage = ex.error_message
            else:
                errorMessage = str(ex)
            self._outputList.Append('S3 Error: "' + errorMessage + '"')

        except Exception as ex:
            errorMessage = str(ex)
            self._outputList.Append('Error: ' + errorMessage)


    def CreateNewFolder(self, newFolderName):
        DEBUG('CreateNewFolder: newFolderName=' + newFolderName)
        try:
            bucket, keyName = self.GetCurrentRemoteLocation()
            if not keyName.startswith('/'):
                keyName = '/' + keyName
            keyName = keyName + newFolderName + '/'
            k = Key(bucket)
            fileName = '.$folder$'
            k.key = keyName + fileName.replace('\\', '/')
            DEBUG('key=' + k.key)
            k.set_contents_from_string('dummy file')

        except S3ResponseError as ex:
            if len(ex.error_message) > 0:
                errorMessage = ex.error_message
            else:
                errorMessage = str(ex)
            self._outputList.Append('S3 Error: "' + errorMessage + '"')

        except Exception as ex:
            errorMessage = str(ex)
            self._outputList.Append('Error: ' + errorMessage)


    def GetNewCryptoFolderName(self):
        if len(self.cryptoDirHoles) > 0:
            newNum = self.cryptoDirHoles[0]
            self.cryptoDirHoles = self.cryptoDirHoles[1:]
        else:
            newNum = self.cryptoDirNextFree
            self.cryptoDirNextFree = self.cryptoDirNextFree + 1

        newName = "_CFD%07d" % (newNum)
        DEBUG('GetNewCryptoFolderName: newName=' + newName)
        return newName

           
    def RemoveExistingCryptoFileName(self, cleartextFileName):
        DEBUG('RemoveExistingCryptoFileName: cleartextFileName=' + cleartextFileName)
        self.cryptoFiles = filter(lambda (s1, s2, s3, s4): s2 != cleartextFileName, self.cryptoFiles)


    def GetAndRemoveExistingCryptoFileName(self, cleartextFileName):
        DEBUG('GetAndRemoveExistingCryptoFileName: cleartextFileName=' + cleartextFileName)
        fileNameList = filter(lambda (s1, s2, s3, s4): s2 == cleartextFileName, self.cryptoFiles)
        self.cryptoFiles = filter(lambda (s1, s2, s3, s4): s2 != cleartextFileName, self.cryptoFiles)
        DEBUG('GetAndRemoveExistingCryptoFileName: cryptoFiles=' + str(self.cryptoFiles))
        if len(fileNameList) == 0:
            DEBUG('GetAndRemoveExistingCryptoFileName: not found')
            return None
        else:
            (s1, s2, s3, s4) = fileNameList[0]
            DEBUG('GetAndRemoveExistingCryptoFileName: found ' + s1)
            return s1


    def GetAndRemoveExistingCryptoDirName(self, cleartextDirName):
        DEBUG('GetAndRemoveExistingCryptoDirName: cleartextDirName=' + cleartextDirName)
        dirNameList = filter(lambda (s1, s2, s3, s4): s2 == cleartextDirName, self.cryptoDirs)
        self.cryptoDirs = filter(lambda (s1, s2, s3, s4): s2 != cleartextDirName, self.cryptoDirs)
        DEBUG('GetAndRemoveExistingCryptoDirName: cryptoDirs=' + str(self.cryptoDirs))
        if len(dirNameList) == 0:
            DEBUG('GetAndRemoveExistingCryptoDirName: not found')
            return None
        else:
            (s1, s2, s3, s4) = dirNameList[0]
            DEBUG('GetAndRemoveExistingCryptoDirName: found ' + s1)
            return s1


    def CryptoDirNameExists(self, cleartextDirName):
        DEBUG('CryptoDirNameExists: cleartextDirName=' + cleartextDirName)
        dirNameList = filter(lambda (s1, s2, s3, s4): s2 == cleartextDirName, self.cryptoDirs)
        if len(dirNameList) > 0:
            DEBUG('CryptoDirNameExists: True')
            return True
        else:
            DEBUG('CryptoDirNameExists: False')
            return False


    def GetCryptoFolder(self, cleartextDirName):
        DEBUG('GetCryptoFolder: cleartextDirName=' + cleartextDirName)
        dirNameList = filter(lambda (s1, s2, s3, s4): s2 == cleartextDirName, self.cryptoDirs)
        if len(dirNameList) == 1:
            (s1, s2, s3, s4) = dirNameList[0]
            DEBUG('GetCryptoFolder: found ' + s1)
            return s1
        else:
            raise Exception('GetCryptoFolder ' + cleartextDirName + ' found ' + str(len(dirNameList)) + ' entries')


    def GetCryptoFileName(self, cleartextFileName):
        DEBUG('GetCryptoFileName: cleartextFileName=' + cleartextFileName)
        fileNameList = filter(lambda (s1, s2, s3, s4): s2 == cleartextFileName, self.cryptoFiles)
        if len(fileNameList) == 1:
            (s1, s2, s3, s4) = fileNameList[0]
            DEBUG('GetCryptoFileName: found ' + s1)
            return s1
        else:
            raise Exception('GetCryptoFileName ' + cleartextFileName + ' found ' + str(len(fileNameList)) + ' entries')


    def GetCleartextFileName(self, cryptoFileName):
        DEBUG('GetCleartextFileName: cryptoFileName=' + cryptoFileName)
        fileNameList = filter(lambda (s1, s2, s3, s4): s1 == cryptoFileName, self.cryptoFiles)
        if len(fileNameList) == 1:
            (s1, s2, s3, s4) = fileNameList[0]
            DEBUG('GetCleartextFileName: found ' + s2)
            return s2
        else:
            raise Exception('GetCleartextFileName ' + cryptoFileName + ' found ' + str(len(fileNameList)) + ' entries')


    def GetCleartextDirName(self, cryptoDirName):
        cryptoDirName = cryptoDirName.rstrip('/')
        DEBUG('GetCleartextDirName: cryptoDirName=' + cryptoDirName)
        dirNameList = filter(lambda (s1, s2, s3, s4): s1 == cryptoDirName, self.cryptoDirs)
        if len(dirNameList) == 1:
            (s1, s2, s3, s4) = dirNameList[0]
            DEBUG('GetCleartextDirName: found ' + s2)
            return s2
        else:
            raise Exception('GetCleartextDirName ' + cryptoDirName + ' found ' + str(len(dirNameList)) + ' entries')


    def GetNewCryptoFileName(self):
        if len(self.cryptoFileHoles) > 0:
            newNum = self.cryptoFileHoles[0]
            self.cryptoFileHoles = self.cryptoFileHoles[1:]
        else:
            newNum = self.cryptoFileNextFree
            self.cryptoFileNextFree = self.cryptoFileNextFree + 1

        newName = "_CFL%07d" % (newNum)
        DEBUG('GetNewCryptoFileName: newName=' + newName)
        return newName

           
    def GetEncryptionKey(self):
        DEBUG('Use pbkdf2_sha256 to generate the key from the password')
        DEBUG('password=' + self._password)
        if (len(self._salt) > 0):
            DEBUG('len(salt)=' + str(len(self._salt)))
            hashAndSalt = pbkdf2_sha256.encrypt(self._password, salt=self._salt, rounds=50000)
        else:
            DEBUG('No salt passed in')
            hashAndSalt = pbkdf2_sha256.encrypt(self._password, salt_size=32, rounds=50000)
        DEBUG('hashAndSalt=' + hashAndSalt)
        # split out the salt and EK
        cryptoStuff = hashAndSalt.split('$')
        salt = cryptoStuff[3]
        while len(salt) % 4 != 0:
            salt = salt + '='
        DEBUG('b64salt=' + salt)
        self._salt = b64decode(salt, CTA_ALTCHARS)
        DEBUG('len(decoded salt)=' + str(len(self._salt)))
        encryptionKey = cryptoStuff[4]
        while len(encryptionKey) % 4 != 0:
            encryptionKey = encryptionKey + '='
        DEBUG('encryptionKey=' + encryptionKey)
        self._encryptionKey = b64decode(encryptionKey, CTA_ALTCHARS)
        DEBUG('len(decoded encryptionKey)=' + str(len(self._encryptionKey)))


    def UploadNewContentsInfo(self):
        DEBUG('UploadNewContentsInfo: len(self._encryptionKey)=' + str(len(self._encryptionKey)))
        if len(self._encryptionKey) == 0 or len(self._salt) == 0:
            self.GetEncryptionKey()

        # Create a file called _CFCI containing list of real and crypto dir names
        dirContentsClearText = FOLDER_CONTENTS_HEADER + '\n' + self._currentRemoteCryptoPath + '\n'
        for cryptoDir in self.cryptoDirs:
            (cryptoDirName, dirName, dirSize, dirTime) = cryptoDir
            DEBUG('cryptoDirName=' + cryptoDirName)
            DEBUG('dirName=' + dirName)
            DEBUG('dirSize=' + dirSize)
            DEBUG('dirTime=' + dirTime)
            dirContentsClearText = dirContentsClearText + str(cryptoDir) + '\n'

        for cryptoFile in self.cryptoFiles:
            (cryptoFileName, fileName, fileSize, fileTime) = cryptoFile
            DEBUG('cryptoFileName=' + cryptoFileName)
            DEBUG('fileName=' + fileName)
            DEBUG('fileSize=' + fileSize)
            DEBUG('fileTime=' + fileTime)
            dirContentsClearText = dirContentsClearText + str(cryptoFile) + '\n'

        DEBUG('dirContentsClearText=' + dirContentsClearText)
        dirContentsCypherStr = EncryptStr(dirContentsClearText, self._encryptionKey)
        DEBUG('dirContentsCypherStr=' + dirContentsCypherStr)

        dirContentsStr = b64encode(self._salt) + '\n' + dirContentsCypherStr
        DEBUG('dirContentsStr=' + dirContentsStr)
        self.CreateNewFile('_CFCI', dirContentsStr)


    def AddedCryptoFile(self, cryptoFileName, newFileName, fileSize, lastModTime):
        DEBUG('AddedCryptoFile: cryptoFileName=' + cryptoFileName)
        DEBUG('AddedCryptoFile: newFileName=' + newFileName)
        DEBUG('AddedCryptoFile: fileSize=' + fileSize)
        DEBUG('AddedCryptoFile: lastModTime=' + lastModTime)
        self.cryptoFiles.append((cryptoFileName, newFileName, fileSize, lastModTime))


    def AddedCryptoDir(self, cryptoFolderName, newFolderName):
        DEBUG('AddedCryptoDir: cryptoFolderName=' + cryptoFolderName)
        DEBUG('AddedCryptoDir: newFolderName=' + newFolderName)
        self.cryptoDirs.append((cryptoFolderName, newFolderName, '0', '0'))


    def OnNewFolder(self, e):
        dlg = NewBucketOrFolderDlg(False, self, wx.ID_ANY, 'Create new folder')
        if dlg.ShowModal() == wx.ID_OK:
            newFolderName = dlg.GetName()
            if len(self._password) > 0:
                cryptoFolderName = self.GetNewCryptoFolderName()
                self.CreateNewFolder(cryptoFolderName)
                self.AddedCryptoDir(cryptoFolderName, newFolderName)
                self.UploadNewContentsInfo()
            else:
                self.CreateNewFolder(newFolderName)

        dlg.Destroy()
        self.RefreshRemoteDir()


    def OnRename(self, e):
        if self.GetSelectedItemCount() == 1:
            oldName = self.GetItemText(self.GetFirstSelected())
            if oldName != None and oldName != '':
                dlg = RenameDlg(oldName, self, wx.ID_ANY, 'Rename')
                if dlg.ShowModal() == wx.ID_OK:
                    newName = dlg.GetNewName()
                    # TODO doit

                dlg.Destroy()
        

    def OnDeleteBucket(self, e):
        prompt = True
        cancelled = False
        currentItem = self.GetFirstSelected()
        while currentItem != -1 and not cancelled:
            bucketName = self.GetItem(currentItem, 0).GetText()
            fileType = self.GetItem(currentItem, 2).GetText()
            if fileType != 'Bucket':
                self._outputList.Append('Expecetd a bucket: ' + fileType)
                currentItem = self.GetNextSelected(currentItem)
                continue

            objectName = bucketName
            action = Utils.ACTION_DELETE_BUCKET
            dlg = Utils.WarningDlg(action, False, objectName, self, wx.ID_ANY, 'Cloud Filer - Warning')
            dlg.ShowModal()
            status = dlg.GetEndId()
            dlg.Destroy()
            if status == wx.ID_OK:

                try:
                    # Delete bucket contents
                    bucket = self.GetBucket(bucketName)
                    keys = bucket.list()
                    # create deletethread(keys)
                    dlg = CloudActionDlg(Utils.ACTION_DELETE_BUCKET, self, wx.ID_ANY, 'Cloud Filer - Delete Bucket')
                    dlg.SetActionMsg(bucketName)
                    thread = DeleteThread(dlg, keys)
                    thread.start()
                    dlg.ShowModal()
                    thread.join()
                    self._outputList.Append(thread.GetMessages())

                    cancelled = dlg.GetCancelled()

                    # If the user didn't cancel and the bucket is empty
                    bucketEmpty = True
                    keys = bucket.list()
                    for key in keys:
                        bucketEmpty = False
                        break

                    if not cancelled:
                        if bucketEmpty:
                            # Delete bucket
                            self._connection.delete_bucket(bucketName)
                            self._outputList.Append('Deleted bucket ' + bucketName)
                        else:
                            self._outputList.Append('Bucket deletion failed')
                    else:
                        self._outputList.Append('Bucket deletion cancelled')

                except S3ResponseError as ex:
                    errorMessage = ex.error_message
                    self._outputList.Append('S3 Error: "' + errorMessage + '"')

                except Exception as ex:
                    errorMessage = str(ex)
                    self._outputList.Append('Error: ' + errorMessage)

    #            dlg.Destroy()

            elif status == wx.ID_CANCEL:
                cancelled = True
                self._outputList.Append('Deletion cancelled')
            else:
                self._outputList.Append('Skipping deletion of ' + bucketName)

            currentItem = self.GetNextSelected(currentItem)

        self.buckets = self._connection.get_all_buckets()
        self.ChangeRemoteDir(self._cloudRootText)


    def RecursiveList(self, bucket, keyName):
        keys = bucket.list(keyName, '/')
        for key in keys:
            DEBUG('key.name=' + key.name)
            if key.name.endswith('/'):
                keys2 = self.RecursiveList(bucket, key.name)
                keys = chain(keys, keys2)

        return keys


    def OnDelete(self, e):
        prompt = True
        cancelled = False
        currentItem = self.GetFirstSelected()
        while currentItem != -1 and not cancelled:
            fileName = self.GetItem(currentItem, 0).GetText()
            fileType = self.GetItem(currentItem, 2).GetText()
            showApplyToAll = False

            if fileType == 'Folder':
                action = Utils.ACTION_DELETE_FOLDER
            else:
                action = Utils.ACTION_DELETE_FILE
            if prompt:
                if self.GetSelectedItemCount() > 1:
                    showApplyToAll = True
                dlg = Utils.WarningDlg(action, showApplyToAll, fileName, self, wx.ID_ANY, 'Cloud Filer - Warning')
                dlg.ShowModal()
                status = dlg.GetEndId()
                dlg.Destroy()
            if status == wx.ID_OK:
                if showApplyToAll and dlg.GetApplyAll():
                    prompt = False
                try:
                    bucket, keyName = self.GetCurrentRemoteLocation()

                    if action == Utils.ACTION_DELETE_FILE:
                        # Just delete the file in main thread
                        if len(self._password) > 0:
                            # Remove file from crypto contents info file
                            savedCleartextFileName = fileName
                            fileName = self.GetAndRemoveExistingCryptoFileName(fileName)
                            self.UploadNewContentsInfo()
                        key = Key(bucket)
                        key.key = keyName + fileName.replace('\\', '/')
                        DEBUG('key=' + key.key)
                        key.delete()
                        if len(self._password) > 0:
                            fileName = savedCleartextFileName
                        else:
                            fileName = key.name
                        self._outputList.Append('Deleted file ' + fileName)
                    else:
                        # Delete folder
                        if len(self._password) > 0:
                            savedCleartextDirName = fileName
                            fileName = self.GetAndRemoveExistingCryptoDirName(fileName)
                            self.UploadNewContentsInfo()
                        keyName = keyName.lstrip('/') + fileName + '/'
                        DEBUG('List folder for deletion: prefix = ' + keyName)
                        keys = self.RecursiveList(bucket, keyName)

                        dlg = CloudActionDlg(Utils.ACTION_DELETE_BUCKET, self, wx.ID_ANY, 'Cloud Filer - Delete Folder')
                        dlg.SetActionMsg(keyName)
                        thread = DeleteThread(dlg, keys)
                        thread.start()
                        dlg.ShowModal()
                        thread.join()
                        self._outputList.Append(thread.GetMessages())

                        cancelled = dlg.GetCancelled()

                        # If the user didn't cancel
                        if not cancelled:
                            if len(self._password) > 0:
                                dirName = savedCleartextDirName + ' (' + keyName + ')'
                            else:
                                dirName = keyName
                            self._outputList.Append('Deleted folder ' + dirName)
                        else:
                            self._outputList.Append('Folder deletion cancelled')

                except S3ResponseError as ex:
                    errorMessage = ex.error_message
                    self._outputList.Append('S3 Error: "' + errorMessage + '"')

                except Exception as ex:
                    errorMessage = str(ex)
                    self._outputList.Append('Error: ' + errorMessage)

    #            dlg.Destroy()

            elif status == wx.ID_CANCEL:
                cancelled = True
                self._outputList.Append('Deletion cancelled')
            else:
                self._outputList.Append('Skipping deletion of ' + fileName)

            currentItem = self.GetNextSelected(currentItem)

        self.RefreshRemoteDir()
        

    def DoubleClick(self, event):
#        newDir = event.GetText()
#        fileType = self.GetItem(event.GetIndex(), 2).GetText()
        
        if self.GetSelectedItemCount() == 1:
            fileName = self.GetItem(self.GetFirstSelected(), 0).GetText()
            fileType = self.GetItem(self.GetFirstSelected(), 2).GetText()
            DEBUG('fileName=' + fileName)
            DEBUG('fileType=' + fileType)
            if fileType == 'Folder' or fileType == 'Bucket':
                DEBUG('it is a folder')
                if self._connected:
                    self.ChangeRemoteDir(fileName)

                else:
                    # Try to do local cd
                    if (self.SetDir(fileName)):
                        cwd = os.getcwd()
                        if self._combo.FindString(cwd) == wx.NOT_FOUND:
                            self._combo.Append(cwd)

                        self._combo.SetStringSelection(cwd)
            else:
                DEBUG('not a folder')


    def WindowsRootDir(self, pathStr):
        if len(pathStr) != 3:
            return False

        return re.match(r'[A-Z]:\\', pathStr)



    def RefreshDir(self):
        self.SetDir(os.getcwd())


#gbl_ignore_pattern = '.*'

    def SetDir(self, dirPath):
        DEBUG('SetDir: ' + dirPath)
        try:
            os.chdir(dirPath)

        except OSError:
            return False

        self.DeleteAllItems()

        self._outputList.Append('lcd ' + dirPath)

        self.displayRoot = os.getcwd() != '/' and not self.WindowsRootDir(os.getcwd())

        files = os.listdir(u'.')

        self.currentFolderFolders = []
        self.currentFolderFiles = []
        # Add folders
        for i in files:
            if os.path.isdir(i):
                DEBUG('isDir: ' + i)
#TODO                if fnmatch.fnmatch(i, gbl_ignore_pattern):
#                    continue

                # Skip any folder we can't access
                try:
                    os.listdir(Utils.ensure_unicode(i))

                except Exception as ex:
                    errorMessage = str(ex)
                    self._outputList.Append('Error: ' + errorMessage)
                    continue  

                size = os.path.getsize(i)
                sec = os.path.getmtime(i)
                timeStr = time.strftime('%Y-%m-%d %H:%M', time.localtime(sec))

                self.currentFolderFolders.append((i, '', 'Folder', timeStr, 1))

        # Add files
        for i in files:
            if not os.path.isdir(i):
#                (name, ext) = os.path.splitext(i)
#TODO                if fnmatch.fnmatch(name, gbl_ignore_pattern):
#                    continue

#                ex = ext[1:].lower()
                
#                fileType, imageIndex = self.GetFileType(ex)

                try:
                    size = os.path.getsize(i)
                    sec = os.path.getmtime(i)
#                    if len(ext) > 0:
#                        name = name + ext

                    timeStr = time.strftime('%Y-%m-%d %H:%M', time.localtime(sec))

                    self.addFileToCurrentFolderList(i, str(size), timeStr)

                except Exception as ex:
                    errorMessage = str(ex)
                    self._outputList.Append('Error: ' + errorMessage)
                    continue  

        self.ResetColumnSorting()
        self.DisplayFileList()
        return True


    def GetFileType(self, ex):

        if (ex == 'gpg' or ex == 'pgp'):
            fileType = 'Encrypted'
            imageIndex = 3
        elif (ex == 'tc'):
            fileType = 'TrueCrypt'
            imageIndex = 3
        elif (ex == 'txt' or ex == 'log'):
            fileType = 'Text'
            imageIndex = 4
        elif (ex == 'nfo'):
            fileType = 'Info'
            imageIndex = 4
        elif (ex == 'pdf'):
            fileType = 'Document'
            imageIndex = 5
        elif (ex == 'png' or ex == 'jpg' or ex == 'bmp' or ex == 'gif'):
            fileType = 'Image'
            imageIndex = 6
        elif (ex == 'tgz' or ex == 'gz' or ex == 'zip' or ex == 'tar' or ex == 'rar' or ex == 'msi' or ex == 'msp'):
            fileType = 'Archive'
            imageIndex = 7
        elif (ex == 'wav' or ex == 'mp3' or ex == 'm4a'):
            fileType = 'Audio'
            imageIndex = 8
        elif (ex == 'avi' or ex == 'mpg' or ex == 'mp4' or ex == 'ogm' or ex =='flv'):
            fileType = 'Video'
            imageIndex = 9
        elif (ex == 'idx' or ex == 'sub' or ex == 'srt'):
            fileType = 'Subtitles'
            imageIndex = 10
        elif (ex == 'iso'):
            fileType = 'Disk Image'
            imageIndex = 11
        elif (ex == 'py'):
            fileType = 'Text'
            imageIndex = 12
        elif (ex == 'odt' or ex == 'doc' or ex == 'docx' or ex == 'docm'):
            fileType = 'Document'
            imageIndex = 13
        elif (ex == 'ods' or ex == 'xls' or ex == 'xlsx' or ex == 'xlsm'):
            fileType = 'Spreadsheet'
            imageIndex = 14
        elif (ex == 'odp' or ex == 'ppt' or ex == 'pptx' or ex == 'pptm' or ex == 'ppsx' or ex == 'ppsm'):
            fileType = 'Presentation'
            imageIndex = 15
        elif (ex == 'odg' or ex == 'dia' or ex == 'vsd' or ex == 'vdx' or ex == 'vsx' or ex == 'vss'):
            fileType = 'Diagram'
            imageIndex = 6
        elif (ex == 'htm' or ex == 'html'):
            fileType = 'Web Page'
            imageIndex = 16
        elif (ex == 'sig' or ex == 'asc'):
            fileType = 'Signature'
            imageIndex = 17
        elif (ex == 'css'):
            fileType = 'Style Sheet'
            imageIndex = Utils.ICON_CSS
        elif (ex == 'js' or ex == 'sh' or ex == 'vbs' or ex == 'pl' or ex == 'ps2' or ex == 'ps1' or ex == 'bat' or ex == 'jse' or ex == 'reg' or ex == 'vb'):
            fileType = 'Script'
            imageIndex = Utils.ICON_SCRIPT
        elif (ex == 'exe'):
            fileType = 'Executable'
            imageIndex = Utils.ICON_EXE
        elif (ex == 'dll' or ex == 'so'):
            fileType = 'Dynamic Library'
            imageIndex = Utils.ICON_DLL
        elif (ex == 'lib' or ex == 'a'):
            fileType = 'Static Library'
            imageIndex = Utils.ICON_LIB
        else:
            fileType = ''
            imageIndex = 2

        return (fileType, imageIndex)


    def S3Connect(self, accessKey, secretKey, password, proxyServer, proxyUsername, proxyPassword, cloudService, cloudHost, cloudRootText):
        try:
            self._connected = False
            self._combo.Clear()
            self._combo.SetValue('')
            self._cryptoDirMap.clear()
            if not self.buckets is None:
                self.buckets = None
            self.DeleteAllItems() # I am a ListCtrl and should be emptied
            self._outputList.Append('Connecting to ' + cloudService + '...')
            if len(proxyServer) > 0:
                parts = proxyServer.split(":")
                if len(parts) != 2:
                    raise Exception("Invalid proxy setting: format should be server:port")

                self._connection = S3Connection(accessKey, secretKey, host=cloudHost, is_secure=True, proxy=parts[0], proxy_port=parts[1], proxy_user=proxyUsername, proxy_pass=proxyPassword)
            else:
                self._connection = S3Connection(accessKey, secretKey, host=cloudHost)

            self.buckets = self._connection.get_all_buckets()
            self._outputList.Append('Connected and found ' + str(len(self.buckets)) + ' buckets')
            self._cloudRootText = cloudRootText
            self._currentRemotePath = self._cloudRootText
            self._currentRemoteCryptoPath = self._cloudRootText

            self._password = password
            self._salt = ''
            if len(password) > 0:
                self.SetBackgroundColour(wx.Colour(200,240,200))
                self._remoteIcon.SetBitmap(wx.Bitmap(Utils.GetFilePath(Utils.ICON_GREEN_SHIELD)))
            else:
                self.SetBackgroundColour(wx.Colour(255,255,255))
                self._remoteIcon.SetBitmap(wx.Bitmap(Utils.GetFilePath(Utils.ICON_RED_SHIELD)))

            self._connected = True

        except S3ResponseError as ex:
            errorMessage = ex.error_message
            if errorMessage is not None:
                self._outputList.Append('S3 Error: "' + errorMessage + '"')
            else:
                self._outputList.Append('S3 Error')

        except Exception as ex:
            errorMessage = str(ex)
            self._outputList.Append('Error: ' + errorMessage)


    def DisplayBucketList(self):
        DEBUG('DisplayBucketList')
        self.currentFolderFolders = []
        self.currentFolderFiles = []
        for i in self.buckets:
            self.currentFolderFolders.append((i.name, '', 'Bucket', '', 18))

        self.displayRoot = False
        self.ResetColumnSorting()
        self.DisplayFileList()


    def RefreshRemoteDir(self):
        self.ChangeRemoteDir(self._currentRemotePath)


    def ChangeRemoteDir(self, newDir):
        DEBUG('ChangeRemoteDir: newDir=' + newDir)
        if not self._connected:
            DEBUG('Not connected: ignore refresh request')
            return

        newDir = newDir.rstrip('/')
        self._outputList.Append('cd ' + newDir)
        # If it's an absolute path
        if newDir.startswith(self._cloudRootText):

            if not self._connected:
                self._outputList.Append('Need to connect first')
                return False

            self._currentRemotePath = newDir

        else:
            # Should be a relatve path and => should have root in _currentRemotePath
            if self._currentRemotePath == None or not self._currentRemotePath.startswith(self._cloudRootText):
                self._outputList.Append('Invalid remote cloud service (expect ' + self._cloudRootText + ' at start) - ' + newDir)

            else:
                if newDir == '..':
                    currentPathBits = self._currentRemotePath.split('/')
                    self._currentRemotePath = '/'.join(currentPathBits[:len(currentPathBits)-1])

                    if len(self._password) > 0:
                        currentCryptoPathBits = self._currentRemoteCryptoPath.split('/')
                        self._currentRemoteCryptoPath = '/'.join(currentCryptoPathBits[:len(currentCryptoPathBits)-1])
                else:
                    self._currentRemotePath = self._currentRemotePath + '/' + newDir
                    if len(self._password) > 0:
                        bucket, bucketListPrefix = self.GetCurrentRemoteLocation()
                        if bucket is not None:
                            self._currentRemoteCryptoPath = self._currentRemoteCryptoPath + '/' + self.GetCryptoFolder(newDir)
                        else:
                            self._currentRemoteCryptoPath = self._currentRemoteCryptoPath + '/' + newDir
        
        self.DisplayRemoteFolder()

        cwd = self._currentRemotePath
        if self._combo.FindString(cwd) == wx.NOT_FOUND:
            self._combo.Append(cwd)
            self._cryptoDirMap[cwd] = self._currentRemoteCryptoPath

        self._combo.SetStringSelection(cwd)

        return True


    def SetRemoteCryptoDirFromCache(self, newDir):
        DEBUG('SetRemoteCryptoDirFromCache: newDir=' + newDir)
        cryptoDir = self._cryptoDirMap[newDir]
        if cryptoDir is None:
            raise Exception('Crypto remote folder not found for ' + newDir)
        self._currentRemoteCryptoPath = cryptoDir


    def GetConnection(self):
        return self._connection


    def GetBucket(self, bucketName):
        for bucket in self.buckets:
            if bucket.name == bucketName:
                return bucket

        raise Exception('Bucket not found: ' + bucketName)


    def GetCurrentRemoteLocation(self):
        if len(self._password) > 0:
            DEBUG('GetCurrentRemoteLocation: currentRemoteCryptoPath=' + self._currentRemoteCryptoPath)
            pathBits = self._currentRemoteCryptoPath.split('/')
        else:
            DEBUG('GetCurrentRemoteLocation: currentRemotePath=' + self._currentRemotePath)
            pathBits = self._currentRemotePath.split('/')
        # If at root: list of buckets
        if len(pathBits) < 2 or pathBits[1] == '':
#            raise Exception('Not in a bucket')
            return None, '/'
        else:
            DEBUG(str(pathBits))
            bucketName = pathBits[1]
            if len(pathBits) > 2:
                bucketListPrefix = '/'.join(pathBits[2:]) + '/'
            else:
                bucketListPrefix = '/'
            return self.GetBucket(bucketName), bucketListPrefix


    def GetEncryptedDirContents(self, bucket, bucketListPrefix):
        # _CFFn or _CFDn
        # FileFreeList is a list of numbers which are holes
        # FileNextFree int
        # get _CFCI which is the contents information file

        self.cryptoDirHoles = []
        self.cryptoFileHoles = []
        self.cryptoDirNextFree = 1
        self.cryptoFileNextFree = 1
        self.cryptoDirs = []
        self.cryptoFiles = []

        self.currentFolderFolders = []
        self.currentFolderFiles = []

        try:
            fileKey = bucketListPrefix + '_CFCI'
            DEBUG('DownloadFile: fileKey = ' + fileKey)
            k = bucket.get_key(fileKey)

            # Seem to have information from S3 service
            if k is None:
                DEBUG('Encrypted contents file not found')
                self.UploadNewContentsInfo() # Create an empty encrypted contents file with a salt
            else:
                DEBUG('key=' + k.key)
                fileSize = k.size
                DEBUG('fileSize=' + str(fileSize))
                
                folderContentsSaltAndCypherStr = k.get_contents_as_string()
                DEBUG('folderContentsSaltAndCypherStr=' + folderContentsSaltAndCypherStr)

                # Split salt from cypher
                splitPoint = folderContentsSaltAndCypherStr.index('\n')
                b64salt = folderContentsSaltAndCypherStr[0:splitPoint]
                folderContentsCypherStr = folderContentsSaltAndCypherStr[splitPoint+1:]

                # Check the salt matches
                DEBUG('b64salt=' + b64salt)
                remoteSalt = b64decode(b64salt)
                if self._salt != remoteSalt:
                    self._salt = remoteSalt
                    self.GetEncryptionKey()

                try:
                    DEBUG('folderContentsCypherStr=' + folderContentsCypherStr)
                    contentsListStr = DecryptStr(folderContentsCypherStr, self._encryptionKey)
                    DEBUG('contentsListStr=' + contentsListStr)

                    # Check the contents info decrypted ok
                    contentsList = contentsListStr.split('\n')
                    if contentsList[0] != FOLDER_CONTENTS_HEADER:
                        raise Exception('Failed to decrypt')
                except Exception as ex:
                    DEBUG('Decryption ex: ' + str(ex))
                    raise Exception('Failed to decrypt')

                DEBUG('currentRemoteCryptoPath=' + self._currentRemoteCryptoPath)
                if self._currentRemoteCryptoPath != contentsList[1]:
                    errorMessage = 'Crypto path mismatch: expected ' + self._currentRemoteCryptoPath + ' found ' + contentsList[1]
                    DEBUG(errorMessage)
                    self._outputList.Append(errorMessage)

                remoteObjList = []
                for i in range(2, len(contentsList) - 1):
#                    (cryptoObjName, objName, objSize, objTime) = literal_eval(contentsList[i])
                    remoteObjList.append(literal_eval(contentsList[i]))

                # Sort the remote list by crypto name so we can work out holes we can reuse and discard duplicate entries
                remoteObjList = sorted(remoteObjList, key=lambda (s1, s2, s3, s4): s1)

                expectedNextDirNumber = 1
                expectedNextFileNumber = 1
                for remoteObj in remoteObjList:
                    (cryptoObjName, objName, objSize, objTime) = remoteObj
                    DEBUG('cryptoObjName=' + cryptoObjName)
                    DEBUG('objName=' + objName)
                    DEBUG('objSize=' + objSize)
                    DEBUG('objTime=' + objTime)
                    objNumber = int(cryptoObjName[4:])
                    if cryptoObjName.startswith('_CFD'):
                        DEBUG('Found a dir')
                        while objNumber > expectedNextDirNumber:
                            self.cryptoDirHoles.append(expectedNextDirNumber)
                            expectedNextDirNumber = expectedNextDirNumber + 1

                        if objNumber != expectedNextDirNumber:
                            errorMessage = 'Duplicate remote dir ref discarded: cryptoObjName=' + cryptoObjName + ' objName=' + objName
                            DEBUG(errorMessage)
                            self._outputList.Append(errorMessage)
                        else:
                            self.AddedCryptoDir(cryptoObjName, objName)
                            expectedNextDirNumber = objNumber + 1
                            self.currentFolderFolders.append((objName, '', 'Folder', '', 18))

                    elif cryptoObjName.startswith('_CFL'):
                        DEBUG('Found a file')
                        while objNumber > expectedNextFileNumber:
                            DEBUG('Found a hole: ' + str(expectedNextFileNumber))
                            self.cryptoFileHoles.append(expectedNextFileNumber)
                            expectedNextFileNumber = expectedNextFileNumber + 1

                        if objNumber != expectedNextFileNumber:
                            errorMessage = 'Duplicate remote file ref discarded: cryptoObjName=' + cryptoObjName + ' objName=' + objName
                            DEBUG(errorMessage)
                            self._outputList.Append(errorMessage)
                        else:
                            self.AddedCryptoFile(cryptoObjName, objName, objSize, objTime)
                            expectedNextFileNumber = objNumber + 1
                            self.addFileToCurrentFolderList(objName, objSize, objTime)
                    else:
                        errorMessage = 'Found a WTF: cryptoObjName=' + cryptoObjName + ' objName=' + objName
                        DEBUG(errorMessage)
                        self._outputList.Append(errorMessage)

                self.cryptoDirNextFree = expectedNextDirNumber
                self.cryptoFileNextFree = expectedNextFileNumber

        except S3ResponseError as ex:
            if ex.error_message != None and len(ex.error_message) > 0:
                msg = ex.error_message
            else:
                msg = str(ex)
            errorMessage = 'S3 Error (getting encrypted dir contents): "' + msg + '"'
            DEBUG(errorMessage)
            self._outputList.Append(errorMessage)

        except Exception as ex:
            errorMessage = 'Error (getting encrypted dir contents): ' + str(ex)
            DEBUG(errorMessage)
            self._outputList.Append(errorMessage)



    def GetRemoteDirContents(self, bucket, bucketListPrefix):
        DEBUG('GetRemoteDirContents: bucket.name=' + bucket.name)
        DEBUG('GetRemoteDirContents: bucketListPrefix=' + bucketListPrefix)
        if bucketListPrefix == '/':
            bucketListPrefix = ''
        files = bucket.list(bucketListPrefix, '/')

        # Seems to have worked so update the folder location
        self.currentFolderFolders = []
        self.currentFolderFiles = []

        # List the folders and files
        for fileEntry in files:
            DEBUG('fileEntry.name=' + fileEntry.name)
            if fileEntry.name == bucketListPrefix + '.$folder$' or fileEntry.name == bucketListPrefix or fileEntry.name.startswith('_CF'):
                continue

            fullName = fileEntry.name[len(bucketListPrefix):]
            if fullName.endswith('/'):
                folderName = fullName.rstrip('/')
                self.currentFolderFolders.append((folderName, '', 'Folder', '', 18))

            else:
                t = time.strptime(fileEntry.last_modified[:19], '%Y-%m-%dT%H:%M:%S')
                timeStr = time.strftime('%Y-%m-%d %H:%M', t)
                self.addFileToCurrentFolderList(fullName, str(fileEntry.size), timeStr)


    def addFileToCurrentFolderList(self, fullName, fileSize, timeStr):
        if '.' in fullName:
            dotIdx = fullName.rfind('.')
            ext = fullName[dotIdx+1:]
        else:
            ext = ''

        ex = ext.lower()
        fileType, imageIndex = self.GetFileType(ex)
        self.currentFolderFiles.append((fullName, fileSize, fileType, timeStr, imageIndex))


    def DisplayRemoteFolder(self):
        DEBUG('Display folder: currentRemotePath=' + self._currentRemotePath)
        DEBUG('Display folder: currentRemoteCryptoPath=' + self._currentRemoteCryptoPath)

        self.DeleteAllItems()

        bucket, bucketListPrefix = self.GetCurrentRemoteLocation()
        if bucket is None:
            self.DisplayBucketList()
        else:
            self._outputList.Append('ls ' + self._currentRemotePath)
            DEBUG('List bucket: prefix = ' + bucketListPrefix)

            if len(self._password) > 0:
                self.GetEncryptedDirContents(bucket, bucketListPrefix)
            else:
                self.GetRemoteDirContents(bucket, bucketListPrefix)

            self.displayRoot = True
            self.ResetColumnSorting()
            self.DisplayFileList()
               

    def DisplayFileList(self):
        if self.sortColumn == SORT_BY_FILENAME:
            # sort by filename
            if self._sortAscending:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s1.lower())
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): s1.lower())
            else:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s1.lower(), reverse=True)
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): s1.lower(), reverse=True)
        elif self.sortColumn == SORT_BY_FILESIZE:
            # sort by filesize
            if self._sortAscending:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s1.lower())
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): int(s2))
            else:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s1.lower())
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): int(s2), reverse=True)
        elif self.sortColumn == SORT_BY_FILETYPE:
            # sort by filetype
            if self._sortAscending:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s1.lower())
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): s3.lower())
            else:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s1.lower())
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): s3.lower(), reverse=True)
        elif self.sortColumn == SORT_BY_MODIFIED:
            # sort by modification timestamp
            if self._sortAscending:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s4)
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): s4)
            else:
                self.currentFolderFolders = sorted(self.currentFolderFolders, key=lambda (s1, s2, s3, s4, i1): s4, reverse=True)
                self.currentFolderFiles = sorted(self.currentFolderFiles, key=lambda (s1, s2, s3, s4, i1): s4, reverse=True)

        if self.displayRoot:
            index = self.InsertStringItem(sys.maxint, '..')
            self.SetStringItem(index, 2, 'Folder')
            self.SetItemImage(index, 0)

        for details in self.currentFolderFolders:
            index = self.InsertStringItem(sys.maxint, details[0])
            self.SetStringItem(index, 1, details[1])
            self.SetStringItem(index, 2, details[2])
            self.SetStringItem(index, 3, details[3])
            self.SetItemImage(index, details[4])

        for details in self.currentFolderFiles:
            index = self.InsertStringItem(sys.maxint, details[0])
            self.SetStringItem(index, 1, details[1])
            self.SetStringItem(index, 2, details[2])
            self.SetStringItem(index, 3, details[3])
            self.SetItemImage(index, details[4])



